"Download files from a public github repo"
# Copyright (c) 2020 Jos Verlinde
# MIT license
# pylint: disable= invalid-name
import os

import requests
from loguru import logger as log

# # log = logging.getLogger(__name__)
# log.setLevel(level=logging.INFO)


def download_file(url: str, module: str, folder: str = "./"):
    "dowload a file from a public github repo"
    local_filename = os.path.abspath(os.path.join(folder, module))
    log.debug("Downloading {:<20} to {}".format(module, local_filename))
    # NOTE the stream=True parameter below
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(local_filename, "wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)
                    # f.flush()
    return local_filename


def download_files(repo, frozen_modules, savepath):
    "dowload multiple files from a public github repo"
    if not os.path.exists(savepath):
        os.makedirs(savepath)

    for mod in frozen_modules:
        url = repo.format(mod)
        download_file(url, mod, savepath)



#!/usr/bin/env python3
"""
Collect modules and python stubs from MicroPython source projects (v1.12 +) and stores them in the all_stubs folder
The all_stubs folder should be mapped/symlinked to the micropython_stubs/stubs repo/folder

"""

# Copyright (c) 2022 Jos Verlinde
# MIT license
# some functions used from micropython/micropython/tools/makemanifest.py,
#   part of the MicroPython project, http://micropython.org/
#   Copyright (c) 2019 Damien P. George


# locating frozen modules :
# tested on MicroPython v1.12 - v1.13
# - 1.16 - using manifests.py, include can specify kwargs
# - 1.13 - using manifests.py, and support for variant
# - 1.12 - using manifests.py, possible also include content of /port/modules folder ?
# - 1.11 and older - include content of /port/modules folder if it exists
import shutil
from pathlib import Path

from loguru import logger as log
from stubber.freeze.common import apply_frozen_module_fixes, get_freeze_path, get_portboard
from stubber.freeze.freeze_manifest_2 import make_path_vars

from .. import utils

# Classes and functions from makemanifest to ensure that the manifest.py files can be processed
from . import makemanifest_1 as makemanifest

# globals
FAMILY = "micropython"

# https://regexr.com/4rh39
# but with an extra P for Python named groups...
regex_port = r"(?P<port>.*[\\/]+ports[\\/]+\w+)[\\/]+boards[\\/]+\w+"  # port
regex_port_board = r"(?P<board>(?P<port>.*[\\/]+ports[\\/]+\w+)[\\/]+boards[\\/]+\w+)"  # port & board


def freeze_one_manifest_1(
    manifest: Path,
    stub_path: Path,
    mpy_path: Path,
    mpy_lib_path: Path,
    version: str,
):
    """
    get and parse the to-be-frozen .py modules for micropython to extract the static type information
    locates the to-be-frozen files through the manifest.py introduced in MicroPython 1.12
    Suitable for version v1.12 -  v1.19
    - manifest.py is used for board specific and daily builds
    - manifest_release.py is used for the release builds
    """

    port, board = get_portboard(manifest)
    path_vars = make_path_vars(port=port, board=board, mpy_path=mpy_path, mpy_lib_path=mpy_lib_path)

    # copy all path vars
    makemanifest.path_vars = path_vars

    # set global for later use - must be an absolute path.
    freeze_path, board = get_freeze_path(stub_path, port, board)

    makemanifest.stub_dir = freeze_path.as_posix()
    # clean target folder
    shutil.rmtree(freeze_path, ignore_errors=True)

    try:
        makemanifest.include(manifest.as_posix())
    except (makemanifest.FreezeError, NameError) as er:
        log.error('freeze error executing "{}": {}'.format(manifest, er.args[0]))

    apply_frozen_module_fixes(freeze_path, mpy_path=mpy_path)

    # make a module manifest
    utils.make_manifest(Path(makemanifest.stub_dir), FAMILY, port=port, board=board, version=version, stubtype="frozen")



from github import Github
from packaging.version import parse


def clean_version(
    version: str,
    *,
    build: bool = False,
    patch: bool = False,
    commit: bool = False,
    drop_v: bool = False,
    flat: bool = False,
):
    "Clean up and transform the many flavours of versions"
    # 'v1.13.0-103-gb137d064e' --> 'v1.13-103'

    if version in ["", "-"]:
        return version
    nibbles = version.split("-")
    if not patch:
        if nibbles[0] >= "1.10.0" and nibbles[0].endswith(".0"):
            # remove the last ".0"
            nibbles[0] = nibbles[0][0:-2]
    if len(nibbles) == 1:
        version = nibbles[0]
    elif build:
        if not commit:
            version = "-".join(nibbles[0:-1])
        else:
            version = "-".join(nibbles)
    else:
        # version = "-".join((nibbles[0], LATEST))
        # HACK: this is not always right, but good enough most of the time
        version = "latest"
    if flat:
        version = version.strip().replace(".", "_")
    else:
        version = version.strip().replace("_", ".")

    if drop_v:
        version = version.lstrip("v")
    else:
        # prefix with `v` but not before latest
        if not version.startswith("v") and version.lower() != "latest":
            version = "v" + version
    return version


def micropython_versions(start="v1.9.2"):
    g = Github()
    repo = g.get_repo("micropython/micropython")
    return [tag.name for tag in repo.get_tags() if parse(tag.name) >= parse(start)]


# def micropython_versions():
#     "static version"
#     return ['v1.19.1', 'v1.19', 'v1.18', 'v1.17', 'v1.16', 'v1.15', 'v1.14', 'v1.13', 'v1.12', 'v1.11', 'v1.10', 'v1.9.4', 'v1.9.3']



"""
get-docstubs

"""

from pathlib import Path

import click
import stubber.basicgit as git
import stubber.utils as utils
from loguru import logger as log
from stubber.stubs_from_docs import generate_from_rst
from stubber.utils.config import CONFIG

from .cli import stubber_cli

##########################################################################################
# log = logging.getLogger("stubber")
#########################################################################################


@stubber_cli.command(name="get-docstubs")
# todo: allow multiple source
@click.option("--path", "-p", default=CONFIG.repo_path.as_posix(), type=click.Path(file_okay=False, dir_okay=True), show_default=True)
@click.option(
    "--stub-path",
    "--stub-folder",
    "target",
    default=CONFIG.stub_path.as_posix(),
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    help="Destination of the files to be generated.",
    show_default=True,
)
@click.option("--family", "-f", "basename", default="micropython", help="Micropython family.", show_default=True)
@click.option("--black/--no-black", "-b/-nb", default=True, help="Run black", show_default=True)
@click.pass_context
def cli_docstubs(
    ctx,
    path: str = CONFIG.repo_path.as_posix(),
    target: str = CONFIG.stub_path.as_posix(),
    black: bool = True,
    basename="micropython",
):
    """
    Build stubs from documentation.

    Read the Micropython library documentation files and use them to build stubs that can be used for static typechecking.
    """

    if path == CONFIG.repo_path.as_posix():
        # default
        rst_path = CONFIG.mpy_path / "docs" / "library"
    elif Path(path).stem == "micropython":
        # path to a micropython repo
        rst_path = Path(path) / "docs" / "library"
    else:
        rst_path = Path(path)  # or specify full path
    v_tag = git.get_tag(rst_path.as_posix())
    if not v_tag:
        # if we can't find a tag , bail
        raise ValueError("No valid Tag found")
    v_tag = utils.clean_version(v_tag, flat=True, drop_v=False)
    release = git.get_tag(rst_path.as_posix(), abbreviate=False) or ""

    dst_path = Path(target) / f"{basename}-{v_tag}-docstubs"

    # get verbose from the parent context
    verbose = ctx.obj["verbose"] != 0

    generate_from_rst(rst_path, dst_path, v_tag, release=release, verbose=verbose, suffix=".pyi")

    # no need to generate .pyi in post processing
    log.info(f"::group:: start post processing of retrieved stubs")
    utils.do_post_processing([dst_path], False, black)
    log.info(f"::group:: Done")



import re
import shutil
from pathlib import Path
from typing import Tuple

from loguru import logger as log


def get_portboard(manifest_path: Path):
    """
    returns a 2-tuple of the port and board in the provided manifest path

    raises an ValueError if neither a port or board can be found
    """
    # https://regex101.com/r/tv7JX4/1
    re_pb = r".*micropython[\/\\]ports[\/\\](?P<port>[\w_-]*)([\/\\]boards[\/\\](?P<board>\w*))?[\/\\]"
    mpy_port = mpy_board = ""
    matches = re.search(re_pb, manifest_path.absolute().as_posix())
    if matches:
        # port and board
        mpy_port = str(matches.group("port") or "")
        mpy_board = str(matches.group("board") or "")
        return mpy_port, mpy_board
    log.error(f"no port or board found in {manifest_path}")
    raise (ValueError("Neither port or board found in path"))


def get_freeze_path(stub_path: Path, port: str, board: str) -> Tuple[Path, str]:
    """
    get path to a folder to store the frozen stubs for the given port/board
    """
    if port == "":
        raise ValueError("port must be provided")

    if board == "":
        board = "GENERIC"

    if board == "manifest_release":
        board = "RELEASE"
    # set global for later use - must be an absolute path.
    freeze_path = (stub_path / port / board).absolute()
    return freeze_path, board


def apply_frozen_module_fixes(freeze_path: Path, mpy_path: Path):
    """
    apply common fixes to the fozen modules to improve stub generation
    """
    # NOTE: FIX 1 add __init__.py to umqtt
    if (freeze_path / "umqtt/robust.py").exists():  # and not (freeze_path / "umqtt" / "__init__.py").exists():
        log.debug(f"add missing : umqtt/__init__.py")
        with open(freeze_path / "umqtt" / "__init__.py", "a") as f:
            f.write("# dummy")

    # NOTE: FIX 2 compensate for expicitly omited task.py from freeze manifest
    # this is normally implemented as a C module, let's use the .py version to generate a stub for this
    if (freeze_path / "uasyncio").exists() and not (freeze_path / "uasyncio" / "task.py").exists():
        # copy task.py from micropython\extmod\uasyncio\task.py to stub_folder
        log.debug(f"add missing : uasyncio/task.py")
        task_py = mpy_path / "extmod" / "uasyncio" / "task.py"
        try:
            shutil.copy(str(task_py), str(freeze_path / "uasyncio"))
        except OSError as er:
            log.warning(f"error copying {task_py} : {er}")
            # try to continue



##########################################################################################
# core
##########################################################################################

from loguru import logger as log
from pathlib import Path
from typing import List

import click
import stubber.get_cpython as get_cpython
import stubber.utils as utils
from stubber.utils.config import CONFIG

from .cli import stubber_cli

##########################################################################################
# log = logging.getLogger("stubber")
#########################################################################################


@stubber_cli.command(name="get-core")
@click.option(
    "--stub-folder",
    "-stubs",
    default=CONFIG.stub_path.as_posix(),
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    show_default=True,
)
@click.option("--pyi/--no-pyi", default=True, help="Create .pyi files for the (new) frozen modules", show_default=True)
@click.option("--black/--no-black", default=True, help="Run black on the (new) frozen modules", show_default=True)
def cli_get_core(
    stub_folder: str = CONFIG.stub_path.as_posix(),
    # core_type: str = "pycopy",  # pycopy or Micropython CPython stubs
    pyi: bool = True,
    black: bool = True,
):
    """
    Download core CPython stubs from PyPi.

    Get the core (CPython compat) modules for both MicroPython and Pycopy.
    """

    stub_paths: List[Path] = []
    for core_type in ["pycopy", "micropython"]:
        log.info(f"::group:: Get Cpython core :{core_type}")
        req_filename = f"requirements-core-{core_type}.txt"
        stub_path = Path(stub_folder) / f"cpython_core-{core_type}"

        get_cpython.get_core(stub_path=stub_path.as_posix(), requirements=req_filename, family=core_type)
        stub_paths.append(stub_path)

    log.info(f"::group:: start post processing of retrieved stubs")
    utils.do_post_processing(stub_paths, pyi, black)
    log.info(f"::group:: Done")



from typing import List, Union

import click
from loguru import logger as log
from stubber.commands.cli import stubber_cli
from stubber.publish.publish import publish_multiple
from tabulate import tabulate

##########################################################################################
# log = logging.getLogger("stubber")
###########

# TODO: make the VERSION AND PORTS dynamic ( move into simpler Config class)
LAST_VERSION = "1.19.1"
ALL_VERSIONS = ["1.17", "1.18", "1.19", "1.19.1"]  # "1.14", "1.15", "1.16","1.17",
ALL_PORTS = ["stm32", "esp32", "esp8266", "rp2"]
ALL_BOARDS = ["GENERIC"]


@stubber_cli.command(name="publish")
@click.option("--family", default="micropython", type=str, show_default=True)
@click.option(
    "--version",
    "--Version",
    "-V",
    "versions",
    multiple=True,
    default=[LAST_VERSION],
    # type=click.Choice(ALL_VERSIONS),
    show_default=True,
    help="multiple: ",
)
@click.option(
    "--port",
    "-p",
    "ports",
    multiple=True,
    default=["auto"],
    #    type=click.Choice(ALL_PORTS),
    show_default=True,
    help="multiple: ",
)
@click.option(
    "--board",
    "-b",
    "boards",
    multiple=True,
    default=["GENERIC"],  # or "auto" ?
    #    type=click.Choice(ALL_BOARDS),
    show_default=True,
    help="multiple: ",
)
@click.option(
    "--pypi/--test-pypi",
    "production",
    is_flag=True,
    default=False,
    show_default=True,
    prompt="Publish to PYPI (y) or Test-PYPI (n)",
    help="publish to PYPI or Test-PYPI",
)
@click.option(
    "--dry-run",
    "dryrun",
    is_flag=True,
    default=False,
    help="go though the motions but do not publish",
)
@click.option(
    "--force",
    is_flag=True,
    default=False,
    help="create new post release even if no changes detected",
)
@click.option(
    "--clean",
    is_flag=True,
    default=False,
    help="clean folders after processing and publishing",
)

# @click.option(
#     "--type",
#     "-t",
#     "stub_type",
#     default=ALL_TYPES[0],
#     show_default=True,
#     type=click.Choice(ALL_TYPES),
#     help="stub type to publish",
# )


def cli_publish(
    family: str,
    versions: Union[str, List[str]],
    ports: Union[str, List[str]],
    boards: Union[str, List[str]],
    production: bool,
    dryrun: bool,
    force: bool,
    clean: bool,
    # stub_type: str,
):
    """
    Commandline interface to publish stubs.
    """
    # force overrules dryrun
    if force:
        dryrun = False
    # lists please
    versions = list(versions)
    ports = list(ports)
    boards = list(boards)

    # db = get_database(publish_path=CONFIG.publish_path, production=production)
    destination = "pypi" if production else "test-pypi"
    log.info(f"Publish {family} {versions} {ports} {boards} to {destination}")

    results = publish_multiple(
        frozen=True,
        family=family,
        versions=versions,
        ports=ports,
        boards=boards,
        production=production,
        dryrun=dryrun,
        force=force,
        clean=clean,
    )
    log.info(tabulate(results, headers="keys"))



#!/usr/bin/env python3
#
# This file is part of the MicroPython project, http://micropython.org/
#
# The MIT License (MIT)
#
# Copyright (c) 2022 Jim Mussared
# Copyright (c) 2019 Damien P. George
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function

import contextlib
import glob
import os
import sys
import tempfile
from collections import namedtuple
from typing import Dict, List, Optional

__all__ = ["ManifestFileError", "ManifestFile"]

# Allow freeze*() etc.
MODE_FREEZE = 1
# Only allow include/require/module/package.
MODE_COMPILE = 2


# In compile mode, .py -> KIND_COMPILE_AS_MPY
# In freeze mode, .py -> KIND_FREEZE_AS_MPY, .mpy->KIND_FREEZE_MPY
KIND_AUTO = 1
# Freeze-mode only, .py -> KIND_FREEZE_AS_MPY, .mpy->KIND_FREEZE_MPY
KIND_FREEZE_AUTO = 2

# Freeze-mode only, The .py file will be frozen as text.
KIND_FREEZE_AS_STR = 3
# Freeze-mode only, The .py file will be compiled and frozen as bytecode.
KIND_FREEZE_AS_MPY = 4
# Freeze-mode only, The .mpy file will be frozen directly.
KIND_FREEZE_MPY = 5
# Compile mode only, the .py file should be compiled to .mpy.
KIND_COMPILE_AS_MPY = 6

# File on the local filesystem.
FILE_TYPE_LOCAL = 1
# URL to file. (TODO)
FILE_TYPE_HTTP = 2


class ManifestFileError(Exception):
    pass


# The set of files that this manifest references.
ManifestOutput = namedtuple(
    "ManifestOutput",
    [
        "file_type",  # FILE_TYPE_*.
        "full_path",  # The input file full path.
        "target_path",  # The target path on the device.
        "timestamp",  # Last modified date of the input file.
        "kind",  # KIND_*.
        "metadata",  # Metadata for the containing package.
        "opt",  # Optimisation level (or None).
    ],
)


# Represent the metadata for a package.
class ManifestMetadata:
    def __init__(self):
        self.version = None
        self.description = None
        self.license = None

    def update(self, description=None, version=None, license=None):
        if description:
            self.description = description
        if version:
            self.version = version
        if license:
            self.license = version


# Turns a dict of options into a object with attributes used to turn the
# kwargs passed to include() and require into the "options" global in the
# included manifest.
#   options = IncludeOptions(foo="bar", blah="stuff")
#   options.foo # "bar"
#   options.blah # "stuff"
class IncludeOptions:
    def __init__(self, **kwargs):
        self._kwargs = kwargs
        self._defaults = {}

    def defaults(self, **kwargs):
        self._defaults = kwargs

    def __getattr__(self, name):
        return self._kwargs.get(name, self._defaults.get(name, None))


class ManifestFile:
    """
    Process a firmware manifest files

    params:
        mode: MODE_*. The mode to process the manifest in, MODE_FREEZE or MODE_COMPILE.
        path_vars: dict of path variables to substitute in the manifest.
            {
                "MPY_DIR": (str) path to the micropython repo,
                "MPY_LIB_DIR": (str) path to the micropython-lib (sub)repo,
                "PORT_DIR": (str) full path to the port directory,
                "BOARD_DIR": (str) full path to the board directory,
                            or the same as PORT_DIR if no port,
                            or the location of a variant's manifest file
            }

    """

    def __init__(self, mode: int, path_vars: Optional[Dict[str, str]] = None):
        # Either MODE_FREEZE or MODE_COMPILE.
        self._mode = mode
        # Path substition variables.
        self._path_vars = path_vars or {}
        # List of files (as ManifestFileResult) references by this manifest.
        self._manifest_files: List[ManifestOutput] = []
        # Don't allow including the same file twice.
        self._visited = set()
        # Stack of metadata for each level.
        self._metadata = [ManifestMetadata()]

    def _resolve_path(self, path):
        # Convert path to an absolute path, applying variable substitutions.
        for name, value in self._path_vars.items():
            if value is not None:
                path = path.replace("$({})".format(name), value)
        return os.path.abspath(path)

    def _manifest_globals(self, kwargs):
        # This is the "API" available to a manifest file.
        return {
            "os": os,
            "metadata": self.metadata,
            "include": self.include,
            "require": self.require,
            "package": self.package,
            "module": self.module,
            "freeze": self.freeze,
            "freeze_as_str": self.freeze_as_str,
            "freeze_as_mpy": self.freeze_as_mpy,
            "freeze_mpy": self.freeze_mpy,
            "options": IncludeOptions(**kwargs),
        }

    def files(self):
        return self._manifest_files

    def execute(self, manifest_file):
        if manifest_file.endswith(".py"):
            # Execute file from filesystem.
            self.include(manifest_file, top_level=True)
        else:
            # Execute manifest code snippet.
            try:
                exec(manifest_file, self._manifest_globals({}))
            except Exception as er:
                raise ManifestFileError("Error in manifest: {}".format(er))

    def _add_file(self, full_path, target_path, kind=KIND_AUTO, opt=None):
        # Check file exists and get timestamp.
        try:
            stat = os.stat(full_path)
            timestamp = stat.st_mtime
        except OSError:
            raise ManifestFileError("cannot stat {}".format(full_path))

        # Map the AUTO kinds to their actual kind based on mode and extension.
        _, ext = os.path.splitext(full_path)
        if self._mode == MODE_FREEZE:
            if kind in (
                KIND_AUTO,
                KIND_FREEZE_AUTO,
            ):
                if ext.lower() == ".py":
                    kind = KIND_FREEZE_AS_MPY
                elif ext.lower() == ".mpy":
                    kind = KIND_FREEZE_MPY
        else:
            if kind != KIND_AUTO:
                raise ManifestFileError("Not in freeze mode")
            if ext.lower() != ".py":
                raise ManifestFileError("Expected .py file")
            kind = KIND_COMPILE_AS_MPY

        self._manifest_files.append(ManifestOutput(FILE_TYPE_LOCAL, full_path, target_path, timestamp, kind, self._metadata[-1], opt))

    def _search(self, base_path, package_path, files, exts, kind, opt=None, strict=False):
        base_path = self._resolve_path(base_path)

        if files:
            # Use explicit list of files (relative to package_path).
            for file in files:
                if package_path:
                    file = os.path.join(package_path, file)
                self._add_file(os.path.join(base_path, file), file, kind=kind, opt=opt)
        else:
            if base_path:
                prev_cwd = os.getcwd()
                os.chdir(self._resolve_path(base_path))

            # Find all candidate files.
            for dirpath, _, filenames in os.walk(package_path or ".", followlinks=True):
                for file in filenames:
                    file = os.path.relpath(os.path.join(dirpath, file), ".")
                    _, ext = os.path.splitext(file)
                    if ext.lower() in exts:
                        self._add_file(
                            os.path.join(base_path, file),
                            file,
                            kind=kind,
                            opt=opt,
                        )
                    elif strict:
                        raise ManifestFileError("Unexpected file type")

            if base_path:
                os.chdir(prev_cwd)  # type: ignore

    def metadata(self, description=None, version=None, license=None):
        """
        From within a manifest file, use this to set the metadata for the
        package described by current manifest.

        After executing a manifest file (via execute()), call this
        to obtain the metadata for the top-level manifest file.
        """

        self._metadata[-1].update(description, version, license)
        return self._metadata[-1]

    def include(self, manifest_path, top_level=False, **kwargs):
        """
        Include another manifest.

        The manifest argument can be a string (filename) or an iterable of
        strings.

        Relative paths are resolved with respect to the current manifest file.

        If the path is to a directory, then it implicitly includes the
        manifest.py file inside that directory.

        Optional kwargs can be provided which will be available to the
        included script via the `options` variable.

        e.g. include("path.py", extra_features=True)

        in path.py:
            options.defaults(standard_features=True)

            # freeze minimal modules.
            if options.standard_features:
                # freeze standard modules.
            if options.extra_features:
                # freeze extra modules.
        """
        if not isinstance(manifest_path, str):
            for m in manifest_path:
                self.include(m)
        else:
            manifest_path = self._resolve_path(manifest_path)
            # Including a directory grabs the manifest.py inside it.
            if os.path.isdir(manifest_path):
                manifest_path = os.path.join(manifest_path, "manifest.py")
            if manifest_path in self._visited:
                return
            self._visited.add(manifest_path)
            if not top_level:
                self._metadata.append(ManifestMetadata())
            with open(manifest_path) as f:
                # Make paths relative to this manifest file while processing it.
                # Applies to includes and input files.
                prev_cwd = os.getcwd()
                os.chdir(os.path.dirname(manifest_path))
                try:
                    exec(f.read(), self._manifest_globals(kwargs))
                except Exception as er:
                    raise ManifestFileError("Error in manifest file: {}: {}".format(manifest_path, er))
                os.chdir(prev_cwd)
            if not top_level:
                self._metadata.pop()

    def require(self, name, version=None, unix_ffi=False, **kwargs):
        """
        Require a module by name from micropython-lib.

        Optionally specify unix_ffi=True to use a module from the unix-ffi directory.
        """
        if self._path_vars["MPY_LIB_DIR"]:
            lib_dirs = ["micropython", "python-stdlib", "python-ecosys"]
            if unix_ffi:
                # Search unix-ffi only if unix_ffi=True, and make unix-ffi modules
                # take precedence.
                lib_dirs = ["unix-ffi"] + lib_dirs

            for lib_dir in lib_dirs:
                for manifest_path in glob.glob(
                    os.path.join(self._path_vars["MPY_LIB_DIR"], lib_dir, "**", name, "manifest.py"),
                    recursive=True,
                ):
                    self.include(manifest_path, **kwargs)
                    return
            raise ValueError("Library not found in local micropython-lib: {}".format(name))
        else:
            # TODO: HTTP request to obtain URLs from manifest.json.
            raise ValueError("micropython-lib not available for require('{}').", name)

    def package(self, package_path, files=None, base_path=".", opt=None):
        """
        Define a package, optionally restricting to a set of files.

        Simple case, a package in the current directory:
            package("foo")
        will include all .py files in foo, and will be stored as foo/bar/baz.py.

        If the package isn't in the current directory, use base_path:
            package("foo", base_path="src")

        To restrict to certain files in the package use files (note: paths should be relative to the package):
            package("foo", files=["bar/baz.py"])
        """
        # Include "base_path/package_path/**/*.py" --> "package_path/**/*.py"
        self._search(base_path, package_path, files, exts=(".py",), kind=KIND_AUTO, opt=opt)

    def module(self, module_path, base_path=".", opt=None):
        """
        Include a single Python file as a module.

        If the file is in the current directory:
            module("foo.py")

        Otherwise use base_path to locate the file:
            module("foo.py", "src/drivers")
        """
        # Include "base_path/module_path" --> "module_path"
        base_path = self._resolve_path(base_path)
        _, ext = os.path.splitext(module_path)
        if ext.lower() != ".py":
            raise ManifestFileError("module must be .py file")
        # TODO: version None
        self._add_file(os.path.join(base_path, module_path), module_path, opt=opt)

    def _freeze_internal(self, path, script, exts, kind, opt):
        if script is None:
            self._search(path, None, None, exts=exts, kind=kind, opt=opt)
        elif isinstance(script, str) and os.path.isdir(os.path.join(path, script)):
            self._search(path, script, None, exts=exts, kind=kind, opt=opt)
        elif not isinstance(script, str):
            self._search(path, None, script, exts=exts, kind=kind, opt=opt)
        else:
            self._search(path, None, (script,), exts=exts, kind=kind, opt=opt)

    def freeze(self, path, script=None, opt=None):
        """
        Freeze the input, automatically determining its type.  A .py script
        will be compiled to a .mpy first then frozen, and a .mpy file will be
        frozen directly.

        `path` must be a directory, which is the base directory to _search for
        files from.  When importing the resulting frozen modules, the name of
        the module will start after `path`, ie `path` is excluded from the
        module name.

        If `path` is relative, it is resolved to the current manifest.py.
        Use $(MPY_DIR), $(MPY_LIB_DIR), $(PORT_DIR), $(BOARD_DIR) if you need
        to access specific paths.

        If `script` is None all files in `path` will be frozen.

        If `script` is an iterable then freeze() is called on all items of the
        iterable (with the same `path` and `opt` passed through).

        If `script` is a string then it specifies the file or directory to
        freeze, and can include extra directories before the file or last
        directory.  The file or directory will be _searched for in `path`.  If
        `script` is a directory then all files in that directory will be frozen.

        `opt` is the optimisation level to pass to mpy-cross when compiling .py
        to .mpy.
        """
        self._freeze_internal(path, script, exts=(".py", ".mpy"), kind=KIND_FREEZE_AUTO, opt=opt)

    def freeze_as_str(self, path):
        """
        Freeze the given `path` and all .py scripts within it as a string,
        which will be compiled upon import.
        """
        self._search(path, None, None, exts=(".py"), kind=KIND_FREEZE_AS_STR)

    def freeze_as_mpy(self, path, script=None, opt=None):
        """
        Freeze the input (see above) by first compiling the .py scripts to
        .mpy files, then freezing the resulting .mpy files.
        """
        self._freeze_internal(path, script, exts=(".py"), kind=KIND_FREEZE_AS_MPY, opt=opt)

    def freeze_mpy(self, path, script=None, opt=None):
        """
        Freeze the input (see above), which must be .mpy files that are
        frozen directly.
        """
        self._freeze_internal(path, script, exts=(".mpy"), kind=KIND_FREEZE_MPY, opt=opt)


# Generate a temporary file with a line appended to the end that adds __version__.
@contextlib.contextmanager
def tagged_py_file(path, metadata):
    dest_fd, dest_path = tempfile.mkstemp(suffix=".py", text=True)
    try:
        with os.fdopen(dest_fd, "w") as dest:
            with open(path, "r") as src:
                contents = src.read()
                dest.write(contents)

                # Don't overwrite a version definition if the file already has one in it.
                if metadata.version and "__version__ =" not in contents:
                    dest.write("\n\n__version__ = {}\n".format(repr(metadata.version)))
        yield dest_path
    finally:
        os.unlink(dest_path)


def main():
    import argparse

    cmd_parser = argparse.ArgumentParser(description="List the files referenced by a manifest.")
    cmd_parser.add_argument("--freeze", action="store_true", help="freeze mode")
    cmd_parser.add_argument("--compile", action="store_true", help="compile mode")
    cmd_parser.add_argument(
        "--lib",
        default=os.path.join(os.path.dirname(__file__), "../lib/micropython-lib"),
        help="path to micropython-lib repo",
    )
    cmd_parser.add_argument("--port", default=None, help="path to port dir")
    cmd_parser.add_argument("--board", default=None, help="path to board dir")
    cmd_parser.add_argument(
        "--top",
        default=os.path.join(os.path.dirname(__file__), ".."),
        help="path to micropython repo",
    )
    cmd_parser.add_argument("files", nargs="+", help="input manifest.py")
    args = cmd_parser.parse_args()

    path_vars = {
        "MPY_DIR": os.path.abspath(args.top) if args.top else None,
        "BOARD_DIR": os.path.abspath(args.board) if args.board else None,
        "PORT_DIR": os.path.abspath(args.port) if args.port else None,
        "MPY_LIB_DIR": os.path.abspath(args.lib) if args.lib else None,
    }

    mode = None
    if args.freeze:
        mode = MODE_FREEZE
    elif args.compile:
        mode = MODE_COMPILE
    else:
        print("Error: No mode specified.", file=sys.stderr)
        exit(1)

    m = ManifestFile(mode, path_vars)  # type: ignore
    for manifest_file in args.files:
        try:
            m.execute(manifest_file)
        except ManifestFileError as er:
            print(er, file=sys.stderr)
            exit(1)
    for f in m.files():
        print(f)


if __name__ == "__main__":
    main()



import sys
from pathlib import Path

import mypy.stubgen as stubgen
from loguru import logger as log
from mypy.errors import CompileError

# default stubgen options
STUBGEN_OPT = stubgen.Options(
    pyversion=(3, 8),  # documentation uses position-only argument indicator which requires 3.8 or higher
    no_import=False,
    include_private=True,
    doc_dir="",
    search_path=[],
    interpreter=sys.executable,
    parse_only=False,
    ignore_errors=True,
    modules=[],
    packages=[],
    files=[],
    output_dir="",
    verbose=True,
    quiet=False,
    export_less=False,
)


def generate_pyi_from_file(file: Path) -> bool:
    """Generate a .pyi stubfile from a single .py module using mypy/stubgen"""

    sg_opt = STUBGEN_OPT
    # Deal with generator passed in
    assert isinstance(file, Path)

    sg_opt.files = [str(file)]
    sg_opt.output_dir = str(file.parent)
    try:
        log.debug(f"Calling stubgen on {str(file)}")
        # TDOD: Stubgen.generate_stubs does not provide a way to return the errors
        # such as `cannot perform relative import`

        stubgen.generate_stubs(sg_opt)
        return True
    except (Exception, CompileError, SystemExit) as e:
        # the only way to know if an error was encountered by generate_stubs
        # TODO: Extract info from e.code or e.args[0] and add that to the manifest ?
        log.warning(e.args[0])
        return False


def generate_pyi_files(modules_folder: Path) -> bool:
    """generate typeshed files for all scripts in a folder using mypy/stubgen"""
    # stubgen cannot process folders with duplicate modules ( ie v1.14 and v1.15 )
    # NOTE: FIX 1 add __init__.py to umqtt
    if (modules_folder / "umqtt/robust.py").exists():  # and not (freeze_path / "umqtt" / "__init__.py").exists():
        log.debug(f"add missing : umqtt/__init__.py")
        with open(modules_folder / "umqtt" / "__init__.py", "a") as f:
            f.write("# dummy")

    modlist = list(modules_folder.glob("**/modules.json"))
    r = True
    if len(modlist) > 1:
        # try to process each module seperatlely
        for mod_manifest in modlist:
            ## generate fyi files for folder
            r = r and generate_pyi_files(mod_manifest.parent)
        return r
    else:  # one or less module manifests
        ## generate fyi files for folder
        log.debug("::group::[stubgen] running stubgen on {0}".format(modules_folder))

        Error_Found = False
        sg_opt = STUBGEN_OPT
        sg_opt.files = [str(modules_folder)]
        sg_opt.output_dir = str(modules_folder)
        try:
            stubgen.generate_stubs(sg_opt)
        except (ValueError, SystemExit) as e:
            # the only way to know if an error was encountered by generate_stubs
            # mypy.errors.CompileError and others ?
            # TODO: Extract info from e.code or e.args[0]
            log.warning(e.args[0])
            Error_Found = True

        if Error_Found:
            # in case of failure ( duplicate module in subfolder) then Plan B
            # - run stubgen on each *.py
            log.debug("::group::[stubgen] Failure on folder, attempt to run stubgen per file")
            py_files = list(modules_folder.rglob("*.py"))
            for py in py_files:
                generate_pyi_from_file(py)
                # todo: report failures by adding to module manifest

        # for py missing pyi:
        py_files = list(modules_folder.rglob("*.py"))
        pyi_files = list(modules_folder.rglob("*.pyi"))

        worklist = pyi_files.copy()
        for pyi in worklist:
            # remove all py files that have been stubbed successfully from the list
            try:
                py_files.remove(pyi.with_suffix(".py"))
                pyi_files.remove(pyi)
            except ValueError:
                log.debug(f"no matching py for : {str(pyi)}")

        # now stub the rest
        # note in some cases this will try a file twice
        for py in py_files:
            r = r and generate_pyi_from_file(py)
            # todo: report failures by adding to module manifest
        return r



"""
Work in Progress
----------------

Tries to determine the return type by parsing the docstring and the function signature
 - if the signature contains a return type --> <something> then that is returned
 - check a lookup dictionary of type overrides, 
    if the functionnae is listed, then use the override
 - use re to find phrases such as:
    - 'Returns ..... '
    - 'Gets  ..... '
 - docstring is joined without newlines to simplify parsing
 - then parses the docstring to find references to known types and give then a rating though a hand coded model ()
 - builds a list return type candidates 
 - selects the highest ranking candidate 
 - the default Type is 'Any'
 

to do:

    - regex :
        - 'With no arguments the frequency in Hz is returned.'
        - 'Get or set' --> indicates overloaded/optional return Union[None|...]
        - add regex for 'Query' ` Otherwise, query current state if no argument is provided. `

    - regex :
        - 'With no arguments the frequency in Hz is returned.'
        - 'Get or set' --> indicates overloaded/optional return Union[None|...]
        - add regex for 'Query' ` Otherwise, query current state if no argument is provided. `

    - try if an Azure Machine Learning works as well 
        https://docs.microsoft.com/en-us/azure/machine-learning/quickstart-create-resources
    - 
"""
# ref: https://regex101.com/codegen?language=python
# https://regex101.com/r/Ni8g2z/2

from loguru import logger as log
import re
from typing import Dict, List, Union

from .lookup import LOOKUP_LIST, NONE_VERBS

# These are shown to import
__all__ = [
    "simple_candidates",
    "compound_candidates",
    "object_candidates",
    "distill_return",
    "return_type_from_context",
    "_type_from_context",  # For testing only
    "TYPING_IMPORT",
]


# logging
# # log = logging.getLogger(__name__)

# all possible Types needed
TYPING_IMPORT: List[str] = [
    "from typing import IO, Any, Callable, Coroutine, Dict, Generator, Iterator, List, NoReturn, Optional, Tuple, Union",
]

# --------------------------------------
# Confidence levels

C_DEFAULT = 0  # Any , the default for all
C_NONE = 0.1 + C_DEFAULT  # better than the default Any
C_BASE = 0.1 + C_NONE  # the Base if a return type has been found

C_STR_NAMES = 0.3

C_GENERIC = 0.6
C_DICT = C_GENERIC
C_TUPLE = C_GENERIC
C_LIST = C_GENERIC
C_BOOL = C_GENERIC
C_FLOAT = C_GENERIC
C_STR = C_GENERIC

C_NONE_RETURN = C_GENERIC
C_OBJECTS = 0.01 + C_GENERIC

C_BYTES = 0.01 + C_GENERIC
C_BYTEARRAY = 0.03 + C_GENERIC
C_INT = 0.03 + C_GENERIC
C_UINT = 0.04 + C_GENERIC
C_ITERATOR = 0.4 + C_GENERIC
C_GENERATOR = 0.4 + C_GENERIC

C_INT_SIZES = 0.5 + C_GENERIC  # better match than bytes and bytearray or object
C_INT_LIKE = 0.5 + C_GENERIC

C_LOOKUP = C_GENERIC + 1

# --------------------------------------
# Weights of the different Lookups
WEIGHT_LOOPUPS = 3.0  # Lookup list weight factor
WEIGHT_RETURN_VAL = 3.0  # Lookup list weight factor
WEIGHT_RETURNS = 1.8  # for Docstring returns
WEIGHT_GETS = 1.5  # For docstring Gets

# --------------------------------------

# base has a confidence that is
BASE = {"type": "Any", "confidence": C_BASE, "match": None}

# --------------------------------------
# Regexes
# --------------------------------------

# all regex matches stop at end of sentence:: . ! ? : ;
# Look for "Return Value: xxxx"
RE_RETURN_VALUE = r"Return value\s?:\s?(?P<return>[^.!?:;]*)"
# Look for Returns , but no 'Information'
RE_RETURN = r"Return(?:s?,?|(?:ing)?)\s(?!information)(?P<return>[^.!?:;]*)"
# Look for gets
RE_GETS = r"Gets?\s(?P<return>[^.!?:;]*)"

# --------------------------------------
# Regex for Literals
# --------------------------------------
RE_LIT_AS_A = r"as a\s?(?P<return>[^.!?:;]*)"
RE_LIT_SENTENCE = r"\s?(?P<return>[^.!?:;]*)"


def dist_rate(i) -> float:
    """"""
    max_len = 150  # must occur in the first 150 chars
    linear = max((max_len - i), 1) / max_len
    # quadratic = max(max_len - (i / 18) ** 4, 1) / max_len
    return linear


WORD_TERMINATORS = ".,!;:?"


def simple_candidates(
    type: str,
    match_string: str,
    keywords: List[str],
    rate: float = 0.5,
    exclude: List[str] = [],
):
    """
    find and rate possible types and confidence weighting for simple types.
    Case sensitive
    """
    candidates = []
    if not any(t in match_string for t in keywords) or any(t in match_string for t in exclude):
        # quick bailout , there are no matches, or there is an exclude
        return []

    #  word matching
    match_words = [w.strip(WORD_TERMINATORS) for w in match_string.split()]
    #  kw =  single word -
    for kw in keywords:
        i = match_string.find(kw)
        if not " " in kw:
            # single keyword
            if not kw in match_words:
                continue
        else:
            # key phrase
            if i < 0:
                continue
        # Assume unsigned are int
        result = BASE.copy()
        result["type"] = type
        result["confidence"] = rate * dist_rate(i)  # OK
        log.trace(f" - found '{kw}' at position {i} with rating {dist_rate(i)}")
        candidates.append(result)
    return candidates


def compound_candidates(
    type: str,
    match_string: str,
    keywords: List[str],
    rate: float = 0.85,
    exclude: List[str] = [],
):
    """
    find and rate possible types and confidence weighting for compound types that can have a subscription.
    Case sensitive
    """
    candidates = []
    if not any(t in match_string for t in keywords) or any(t in match_string for t in exclude):
        # quick bailout , there are no matches, or there is an exclude
        return []

    #  word matching
    match_words = [w.strip(WORD_TERMINATORS) for w in match_string.split()]
    #  kw =  single word -
    for kw in keywords:
        i = match_string.find(kw)
        if not " " in kw:
            # single keyword
            if not kw in match_words:
                continue
        else:
            # key phrase
            if i < 0:
                continue
        # List / Dict / Generator of Any / Tuple /
        sub = None
        result = BASE.copy()
        confidence = rate
        for element in ("tuple", "string", "unsigned", "int"):
            if element in match_string.casefold():
                j = match_string.find(element)
                if i == j:
                    # do not match on the same main and sub
                    continue
                confidence += 0.10  # boost as we have a subtype
                if element == "tuple":
                    sub = "Tuple"
                    break
                elif element == "string":
                    sub = "str"
                    break
                elif element == "unsigned":
                    sub = "int"
                    break
                else:
                    sub = element
        if sub:
            result["type"] = f"{type}[{sub}]"
        else:
            result["type"] = f"{type}"
        confidence = confidence * dist_rate(i)  # distance weighting
        result["confidence"] = confidence
        log.trace(f" - found '{kw}' at position {i} with confidence {confidence} rating {dist_rate(i)}")

        candidates.append(result)
    return candidates


def object_candidates(
    match_string: str,
    rate: float = 0.81,
    exclude: List[str] = ["IRQ"],
):
    """
    find and rate possible types and confidence weighting for Object types.
    Case sensitive
    """

    candidates = []
    keywords = [
        "Object",
        "object",
    ]  # Q&D

    if not any(t in match_string for t in keywords) or any(t in match_string for t in exclude):
        # quick bailout , there are no matches, or there is an exclude
        return []
    for kw in keywords:
        i = match_string.find(kw)
        if i < 0:
            continue
        # List / Dict / Generator of Any / Tuple /
        confidence = rate

        # did the word actually occur, or is it just a partial
        words = match_string.split(" ")  # Return <multiple words object>
        if kw in words:
            pos = words.index(kw)
            if pos == 0:
                object = "Any"
            else:
                object = words[pos - 1]
            if object in ("stream-like", "file"):
                object = "IO"  # needs from typing import IO
            elif object == "callback":
                object = "Callable[..., Any]"  # requires additional 'from typing import Callable'
            else:
                # clean
                object = re.sub(r"[^a-z.A-Z0-9]", "", object)
            result = BASE.copy()
            result["type"] = object
            if object in ["an", "any"]:  # "Return an / any object"
                result["type"] = "Any"
                confidence += 0.10  # abstract , but very good
            elif object[0].islower():
                confidence -= 0.20  # not so good
            result["confidence"] = confidence * dist_rate(i)
            candidates.append(result)
    return candidates


def has_none_verb(docstr: str) -> List:
    "returns a None result if the docstring starts with a verb that indicates None"
    docstr = docstr.strip().casefold()
    if any(docstr.startswith(kw.casefold()) for kw in NONE_VERBS):
        result = BASE.copy()
        result["type"] = "None"
        result["confidence"] = C_NONE  # better than the default Any
        return [result]
    else:
        return []


def distill_return(return_text: str) -> List[Dict]:
    """Find return type and confidence.
    Returns a list of possible types and confidence weighting.
    {

        type :str               # the return type
        confidence: float       # the confidence between 0.0 and 1
        match: Optional[str]    # for debugging : the reason the match was made

    }

    """
    candidates = [BASE]  # Default to the base , which is 'Any'

    # clean up match_string
    match_string = return_text.strip().rstrip(".")
    match_string = match_string.replace("`", "")

    candidates += compound_candidates("Generator", match_string, ["generator"], C_GENERATOR)
    candidates += compound_candidates("Iterator", match_string, ["iterator"], C_ITERATOR)
    candidates += compound_candidates("List", match_string, ["a list of", "list of", "an array"], C_LIST)

    candidates += simple_candidates("Dict", match_string, ["a dictionary", "dict", "Dictionary"], C_DICT)
    candidates += simple_candidates(
        "Tuple",
        match_string,
        [
            "tuple",
            "a pair",
            "1-tuple",
            "2-tuple",
            "3-tuple",
            "4-tuple",
            "5-tuple",
            "6-tuple",
            "7-tuple",
            "8-tuple",
            "9-tuple",
        ],
        C_TUPLE,
    )

    candidates += simple_candidates("int", match_string, ["unsigned integer", "unsigned int", "unsigned"], C_UINT)

    candidates += simple_candidates(
        "int",
        match_string,
        [
            "number",
            "integer",
            "count",
            "int",
            "0 or 1",
        ],
        C_INT,
    )

    # good but nor perfect indicators of integers
    # better match than bytes and bytearray or object
    candidates += simple_candidates(
        "int",
        match_string,
        [
            "length",
            "total size",
            "size of",
            "the index",
            "number of",
            "address of",
            "the duration",
        ],
        C_INT_SIZES,
    )

    candidates += simple_candidates("int", match_string, [], C_INT_SIZES)

    # Assume numbers are signed int
    candidates += simple_candidates(
        "int",
        match_string,
        [
            "index",
            "**signed** value",
            "seconds",
            "nanoseconds",
            "millisecond",
            "offset",
        ],
        C_INT_LIKE,
    )

    # better match than bytes
    candidates += simple_candidates("bytearray", match_string, ["bytearray"], C_BYTEARRAY)

    # OK, better than just string
    candidates += simple_candidates("bytes", match_string, ["bytes", "byte string"], C_BYTES)

    candidates += simple_candidates("bool", match_string, ["boolean", "bool", "True", "False"], C_BOOL)
    candidates += simple_candidates(
        "float",
        match_string,
        [
            "float",
            "logarithm",
            "sine",
            "cosine",
            "tangent",
            "exponential",
            "complex number",
            "phase",
            "ratio of",
        ],
        C_FLOAT,
    )

    candidates += simple_candidates("str", match_string, ["string", "(sub)string", "sub-string", "substring"], C_STR)

    candidates += simple_candidates("str", match_string, ["name", "names"], C_STR_NAMES)
    ## "? contains 'None if there is no'  --> Union[Null, xxx]"
    candidates += simple_candidates(
        "None",
        match_string,
        ["``None``", "None"],
        C_NONE_RETURN,
        exclude=["previous value", "if there is no"],
    )

    candidates += object_candidates(match_string, C_OBJECTS)

    return candidates


def return_type_from_context(*, docstring: Union[str, List[str]], signature: str, module: str, literal: bool = False):
    try:
        return str(_type_from_context(module=module, signature=signature, docstring=docstring, literal=literal)["type"])
    except Exception:
        return "Any"


def _type_from_context(*, docstring: Union[str, List[str]], signature: str, module: str, literal: bool = False) : #-> Dict[str , Union[str,float]]:
    """Determine the return type of a function or method based on:
     - the function signature
     - the terminology used in the docstring

    Logic:
    - if the signature contains a return type --> <something> then that is returned
    - use re to find phrases such as:

        - 'Returns ..... '
        - 'Gets  ..... '

    - docstring is joined without newlines to simplify parsing
    - then parses the docstring to find references to known types and give then a rating though a hand coded model ()
    - builds a list return type candidates
    - selects the highest ranking candidate
    - the default Type is 'Any'
    """

    if isinstance(docstring, list):
        # join with space to avoid ending at a newline
        docstring = " ".join(docstring)

    # give the regex that searches for returns a 0.2 boost as that is bound to be more relevant

    if not literal:
        weighted_regex = [
            (RE_RETURN_VALUE, WEIGHT_RETURN_VAL),
            (RE_RETURN, WEIGHT_RETURNS),
            (RE_GETS, WEIGHT_GETS),
            #       (reads_regex, 1.0),
        ]
    else:
        weighted_regex = [
            (RE_LIT_AS_A, 1.0),
            (RE_LIT_SENTENCE, 2.0),
        ]

    # only the function name without the leading module
    function_re = re.compile(r"[\w|.]+(?=\()")

    # matches: List[re.Match] = []
    candidates: List[Dict] = [{"match": "default", "type": "Any", "confidence": 0}]

    # if the signature contains a return type , then use that and do nothing else.
    if "->" in signature:
        sig_type = signature.split("->")[-1].strip(": ")
        return {"type": sig_type, "confidence": WEIGHT_LOOPUPS, "match": signature}

    # ------------------------------------------------------
    # lookup returns that cannot be found based on the docstring from the lookup list
    try:
        function_name = function_re.findall(signature)[0]
    except IndexError:
        function_name = signature.strip().strip(":()")

    function_name = ".".join((module, function_name))

    if function_name in LOOKUP_LIST.keys():
        sig_type = LOOKUP_LIST[function_name][0]
        return {
            "type": sig_type,
            "confidence": C_LOOKUP * WEIGHT_LOOPUPS,
            "match": function_name,
        }
    # ------------------------------------------------------
    # parse the docstring for simple start verbs,
    # and add them as a candidate
    candidates += has_none_verb(docstring)

    # ------------------------------------------------------
    # parse the docstring for the regexes and weigh the results accordingly
    for weighted in weighted_regex:

        match_iter = re.finditer(weighted[0], docstring, re.MULTILINE | re.IGNORECASE)
        for match in match_iter:
            # matches.append(match)
            distilled = distill_return(match.group("return"))
            for item in distilled:
                candidate = {
                    "match": match,
                    "type": item["type"],
                    "confidence": item["confidence"] * weighted[1],  # add search boost
                }
                candidates.append(candidate)
    # Sort
    candidates = sorted(candidates, key=lambda x: x["confidence"], reverse=True)
    best = candidates[0]  # best candidate

    # ref: https://docs.python.org/3/library/typing.html#typing.Coroutine
    # Coroutine[YieldType, SendType, ReturnType]
    # todo: sanity check against actual code .....
    if "This is a coroutine" in docstring and not "Coroutine" in str(best["type"]):  # type: ignore
        best["type"] = f"Coroutine[{best['type']}, Any, Any]"

    # return the best candidate, or Any
    return best  # best candidate



##########################################################################################
# minify
##########################################################################################
from pathlib import Path
from typing import Union

import click
from loguru import logger as log
from stubber.minify import minify

from .cli import stubber_cli


@stubber_cli.command(name="minify")
@click.option(
    "--source", "-s", default="board/createstubs.py", type=click.Path(exists=True, file_okay=True, dir_okay=False), show_default=True
)
@click.option("--target", "-t", default="./minified", type=click.Path(exists=True, file_okay=True, dir_okay=True), show_default=True)
@click.option("--diff", "-d", help="Show the functional changes made to the source script.", default=False, is_flag=True)
@click.option("--compile", "-c", "-xc", "cross_compile", help="Cross compile after minification.", default=False, is_flag=True)
@click.option("--all", "-a", help="Minify all variants (normal, _mem and _db).", default=False, is_flag=True)
@click.option(
    "--report/--no-report",
    "keep_report",
    help="Keep or disable minimal progress reporting in the minified version.",
    default=True,
    show_default=True,
)
@click.pass_context
def cli_minify(
    ctx,
    source: Union[str, Path],
    target: Union[str, Path],
    keep_report: bool,
    diff: bool,
    cross_compile: bool,
    all: bool,
) -> int:
    """
    Minify createstubs*.py.

    Creates a minified version of the SOURCE micropython file in TARGET (file or folder).
    The goal is to use less memory / not to run out of memory, while generating Firmware stubs.
    """
    if all:
        sources = ["board/createstubs.py", "board/createstubs_mem.py", "board/createstubs_db.py"]
    else:
        sources = [source]
    log.trace(f"sources: {sources}")
    for source in sources:
        log.info(f"\nMinifying {source}...")
        minify(source, target, keep_report, diff, cross_compile)  # noqa

    log.info("\nDone!")
    return 0



"""Pre/Post Processing for createstubs.py"""
import subprocess
from pathlib import Path
from typing import List

from loguru import logger as log

from .stubmaker import generate_pyi_files

# # log = logging.getLogger(__name__)


def do_post_processing(stub_paths: List[Path], pyi: bool, black: bool):
    "Common post processing"
    for pth in stub_paths:
        if pyi:
            log.debug("Generate type hint files (pyi) in folder: {}".format(pth))
            generate_pyi_files(pth)
        if black:
            run_black(pth)


def run_black(path: Path, capture_output=False):
    """
    run autoflake to remove unused imports
    needs to be run BEFORE black otherwise it does not recognize long import from`s.
    """
    cmd = [
        "black",
        path.as_posix(),
    ]
    # subprocess.run(cmd, capture_output=log.level >= logging.INFO)
    result = subprocess.run(cmd, capture_output=capture_output)
    return result.returncode


def run_autoflake(path: Path, capture_output=False):
    """
    run autoflake to remove unused imports
    needs to be run BEFORE black otherwise it does not recognize long import from`s.
    """
    cmd = [
        "autoflake",
        "-r",
        "--in-place",
        path.as_posix(),
        "-v",
        "-v",  # show some feedback
    ]
    # subprocess.run(cmd, capture_output=log.level >= logging.INFO)
    result = subprocess.run(cmd, capture_output=capture_output)

    return result.returncode



import os
import shutil
from pathlib import Path
from typing import List, Optional

from loguru import logger as log
from stubber import utils
from stubber.tools.manifestfile import (MODE_FREEZE, ManifestFile,
                                        ManifestFileError, ManifestOutput)
from stubber.utils.config import CONFIG

from .common import apply_frozen_module_fixes, get_freeze_path, get_portboard


def make_path_vars(
    *,
    mpy_path: Path = CONFIG.mpy_path,
    mpy_lib_path: Path = CONFIG.mpy_lib_path,  # ? if <= 1.19.1
    port: Optional[str] = None,
    board: Optional[str] = None,
):
    if port is None or port == "":  # pragma: no cover
        port_path = mpy_path
    else:
        port_path = mpy_path / "ports" / port

    if board is None or board == "":  # pragma: no cover
        board_path = port_path
    else:
        board_path = port_path / "boards" / board

    log.trace(f"port_path : {port_path}")
    log.trace(f"board_path: {board_path}")
    if not port_path.exists():  # pragma: no cover
        raise ValueError("port board path not found")
    if not board_path.exists():  # pragma: no cover
        raise ValueError("board path not found")

    # VARS must be absolute paths
    path_vars = {
        "MPY_DIR": mpy_path.absolute().as_posix(),
        "MPY_LIB_DIR": mpy_lib_path.absolute().as_posix(),
        "PORT_DIR": port_path.absolute().as_posix(),
        "BOARD_DIR": board_path.absolute().as_posix(),
    }
    return path_vars


def freeze_one_manifest_2(manifest: Path, frozen_stub_path: Path, mpy_path: Path, mpy_lib_path: Path, version: str):
    # apparently there can be multiple manifest files to a board ?
    # save cwd for 'misbehaving' older esp8266 manifest files
    cwd = Path.cwd()
    # so we need to get the port and board from the path
    log.info(f"input_manifest: {manifest}")
    port, board = get_portboard(manifest)
    log.info(f"port-board: '{port}-{board}'")

    path_vars = make_path_vars(port=port, board=board, mpy_path=mpy_path, mpy_lib_path=mpy_lib_path)
    upy_manifest = ManifestFile(MODE_FREEZE, path_vars)
    try:
        # assume manifestneeds to be run from the port's folder
        os.chdir(path_vars["PORT_DIR"])
        upy_manifest.execute(manifest.as_posix())
    except ManifestFileError as er:
        log.error('freeze error executing "{}": {}'.format(manifest, er.args[0]))
        raise er
    log.info(f"total {len(upy_manifest.files())} files")

    # restore working directory
    os.chdir(cwd)
    # save the frozen files to the stubs
    copy_frozen_to_stubs(frozen_stub_path, port, board, upy_manifest.files(), version, mpy_path=mpy_path)


def copy_frozen_to_stubs(stub_path: Path, port: str, board: str, files: List[ManifestOutput], version: str, mpy_path: Path):
    """
    copy the frozen files from the manifest to the stubs folder

    stubpath = the destination : # stubs/{family}-{version}-frozen
    """
    freeze_path, board = get_freeze_path(stub_path, port, board)

    log.info(f"copy frozen: {port}-{board} to {freeze_path}")
    freeze_path.mkdir(parents=True, exist_ok=True)
    # clean target folder
    shutil.rmtree(freeze_path, ignore_errors=True)

    # print(tabulate(files))
    # copy the frozen files to the stubs
    for f in files:
        dest = freeze_path / f.target_path
        log.trace(f"copying {f.full_path} to {f.target_path}")
        dest.parent.mkdir(parents=True, exist_ok=True)
        try:
            shutil.copy(f.full_path, dest)
        except OSError as er:
            log.warning(f"error copying {f.full_path} to {dest}: {er}")
            raise er
            # try to continue

    apply_frozen_module_fixes(freeze_path, mpy_path=mpy_path)

    # make a module manifest
    FAMILY = "micropython"
    utils.make_manifest(freeze_path, FAMILY, port=port, board=board, version=version, stubtype="frozen")



"""
this is an list with manual overrides for function returns that could not efficiently be determined 
from their docstring description 
Format: a dictionary with :
- key = module.[class.]function name
- value : two-tuple with ( return type , priority )

"""
# These are shown to import
__all__ = [
    "LOOKUP_LIST",
    "NONE_VERBS",
    "CHILD_PARENT_CLASS",
    "PARAM_FIXES",
    "MODULE_GLUE",
    "RST_DOC_FIXES",
    "DOCSTUB_SKIP",
    "U_MODULES",
]


# modules documented with base name only
U_MODULES = ["os", "time", "array", "binascii", "io", "json", "select", "socket", "ssl", "struct", "zlib"]


# Some classes are documented as functions
# This table is used to try to correct the errors in the documentation.
# it is applied to each .rst file after loading the contents.

RST_DOC_FIXES = [
    # ------------------------------------------------------------------------------------------------
    # re.rst - function and class with the same name
    # todo: issue https://github.com/micropython/micropython/issues/8273
    (".. method:: match.", ".. method:: Match."),
    ("            match.end", "            Match.end"),
    # ------------------------------------------------------------------------------------------------
    # collections.rst - should be fixed in v1.19
    # PR: https://github.com/micropython/micropython/pull/7976
    # keep around for older docstubs
    (".. function:: deque(", ".. class:: deque("),
    (".. function:: OrderedDict(", ".. class:: OrderedDict("),
    # ------------------------------------------------------------------------------------------------
    # generator functions - WILL_NOT_FIX in the MicroPython documentation
    # these are documented as functions, but return an object with the the same name as the function.
    # for static type analysis this is best considered as a Class, so morph them before processing.
    # uselect.rst
    (".. function:: poll(", ".. class:: poll("),
]

# docstubs generation, exclude stub generation for below stubs.
DOCSTUB_SKIP = [
    "uasyncio.rst",  # can create better stubs from frozen python modules.
    "builtins.rst",  # conflicts with static type checking , has very little information anyway
    "re.rst",  # regex is too complex
]

# contains return types for functions and methods that are not clearly documented.
LOOKUP_LIST = {
    "builtins.bytes": ("bytes", 0.95),
    "builtins.from_bytes": ("int", 0.95),
    "builtins.to_bytes": ("bytes", 0.95),
    "bytearray_at": ("bytearray", 0.95),
    "gc.collect": ("None", 0.95),
    "machine.deepsleep": ("NoReturn", 0.95),
    "machine.reset_cause": ("int", 0.95),
    "machine.reset": ("NoReturn", 0.95),  # never returns
    "machine.Signal.value": ("int", 0.95),
    "machine.soft_reset": ("NoReturn", 0.95),  # never returns
    "machine.UART.irq": ("Any", 0.95),  # no IRQ type defined
    "math.isnan": ("bool", 0.95),
    "micropython.opt_level": ("Any", 0.95),  # Not clear in docstring
    "micropython.const": ("int", 0.95),  # const is always an int
    "pyb.hard_reset": ("NoReturn", 0.95),  # never returns
    "pyb.I2C.recv": ("bytes", 0.95),  # complex in docstring
    "pyb.SPI.recv": ("bytes", 0.95),  # complex in docstring
    "ubluetooth.BLE.irq": ("Any", 0.95),  # never returns
    "uctypes.bytearray_at": ("bytearray", 0.95),
    "uctypes.bytes_at": ("bytes", 0.95),
    "uio.open": ("IO", 0.95),  #  Open a file.
    "uos.listdir": ("List[Any]", 0.95),
    "ussl.ussl.wrap_socket": ("IO", 0.95),  # undocumented class ssl.SSLSocket
    "usys.exit": ("NoReturn", 0.95),  # never returns
    "utime.sleep_ms": (
        "Coroutine[None, None, None]",  # Micropython V1.15+ ?
        0.95,
    ),
    "stm.mem8": ("bytearray", 0.95),  # Read/write 8 bits of memory.
    "stm.mem16": ("bytearray", 0.95),  # Read/write 16 bits of memory.
    "stm.mem32": ("bytearray", 0.95),  # Read/write 32 bits of memory.
}


# if no type has been determined, and the docstring starts with one of these verbs, then assume the return type is None
# - The starting word or words with a training space
NONE_VERBS = [
    "Activate ",
    "Build a ",
    "Cancel ",
    "Clear ",
    "Close ",
    "cancel ",
    "Configure ",
    "Connect ",
    "Deactivate ",
    "De-initialises ",
    "Deinitialises ",
    "Delay ",
    "Disable ",
    "Display ",
    "Disconnect ",
    "Draw ",
    "Enable ",
    "Feed the ",
    "Fill the ",
    "Generate ",
    "Initialise the ",
    "Initialize ",  # US/UK spelling
    "Issue a ",
    "Load ",
    "Modify ",
    "Print ",
    "Register ",
    "Remove ",
    "Rename ",
    "Reset ",
    "Resets ",
    "Send ",
    "Sends ",
    "Set Pin ",
    "Set the ",
    "Sets",
    "Show ",
    "Stop ",
    "Stops ",
    "Sync ",
    "Turn ",
    "Wait ",
    "Write ",
    "Writes ",
]

# Add additional imports to generated modules to allow one module te refer to another,
# or to import other supporting modules
MODULE_GLUE = {
    "lcd160cr": ["from .machine import SPI"],  # module returns SPI objects defined in machine
    "esp32": ["from __future__ import annotations"],  # Class methods return Class
    "collections": ["from queue import Queue"],  # dequeu is a subclass
}

# manual fixes needed for parameters ( micropython v.16 & v1.17)
PARAM_FIXES = [
    ("\\*", "*"),  # change weirdly written wildcards \* --> *
    (r"\**", "*"),  # change weirdly written wildcards \* --> *
    (r"/*", "*"),  # change weirdly written wildcards \* --> *
    # ("**", "*"),  # change weirdly written wildcards \* --> *
    ("'param'", "param"),  # loose notation in documentation
    (
        "(adcx, adcy, ...), (bufx, bufy, ...)",
        "adcs, bufs",
    ),  # method:: ADC.read_timed_multi((adcx, adcy, ...), (bufx, bufy, ...), timer)
    (
        "(ip, subnet, gateway, dns)",
        "configtuple",
    ),  # network: # .. method:: AbstractNIC.ifconfig([(ip, subnet, gateway, dns)])
    (
        "(buttons, x, y, z)",
        "hidtuple",
    ),  # pyb  .. function:: hid((buttons, x, y, z))
    (
        "(hostname, port, lambda)",
        "tuple",
    ),  # esp v1.15.2 .. function:: getaddrinfo((hostname, port, lambda))
    # (
    #     "cert_reqs=CERT_NONE",
    #     "cert_reqs=None",
    # ),  # .. function:: ussl.wrap_socket(sock, server_side=False, keyfile=None, certfile=None, cert_reqs=CERT_NONE, ca_certs=None, do_handshake=True)
    (
        "='dhcp' or configtuple: Optional[Any]=None",
        ": Union[str,Tuple]='dhcp'",
    ),  # network.rst method:: WLANWiPy.ifconfig(if_id=0, config=['dhcp' or configtuple])
    (
        "'pgm')",
        "cmd:str ,/)",
    ),  # network.rst .. method:: CC3K.patch_program('pgm')
    (
        "block_device or path",
        "block_device_or_path",
    ),  #
    (
        "(ip, subnet, gateway, dns):Optional[Any]=None",
        "configtuple: Optional[Tuple]",
    ),  # ifconfig
    (
        "lambda",
        "lambda_fn",
    ),  # illegal keywords
    (
        "stride=width",
        "stride=-1",
    ),  # # unresolvable parameter defaults # FrameBuffer: def __init__
    (
        "trigger=(IRQ_FALLING | IRQ_RISING)",
        "trigger=IRQ_FALLING ",
    ),  ## fixes for machine.py class constants
    (
        "pins=(SCK, MOSI, MISO)",
        "pins:Optional[Tuple]",
    ),  #
    ## rp2.PIO.irq
    (
        "trigger=IRQ_SM0|IRQ_SM1|IRQ_SM2|IRQ_SM3",
        "trigger=IRQ_SM0",
    ),
    # SPI.INIT - to fix error: Non-default argument follows default argument
    # fixed in doc v1.18+
    (
        "prescaler, polarity=1",
        "prescaler=1, polarity=1",
    ),
    # machine.Pin.__init__ constructor - Defaults assumed from the documentation.
    # fixed in doc v1.18+
    (
        ", value, drive, alt",
        ", value=None, drive=0, alt=-1",
    ),
    # network.LAN.init
    # def __init__(self, id, *, phy_type=<board_default>, phy_addr=<board_default>, phy_clock=<board_default>) -> None:
    ("=<board_default>", "=0"),
]

# List of classes and their parent classes that should be added to the class definition
CHILD_PARENT_CLASS = {
    # machine
    "SoftSPI": "SPI",
    "SoftI2C": "I2C",
    "Switch": "Pin",
    # uio # unclear regarding deprecation in python 3.12
    "FileIO": "IO",
    "textIOWrapper": "IO",
    "StringIO": "IO",
    "BytesIO": "IO",
    # uzlib
    "DecompIO": "IO",  # https://docs.python.org/3/library/typing.html#other-concrete-types
    # network
    "WLAN": "AbstractNIC",
    "WLANWiPy": "AbstractNIC",
    # uhashlib
    "md5": "hash",
    "sha1": "hash",
    "sha265": "hash",
    # collections
    "OrderedDict": "dict",
    "namedtuple": "tuple",
    "deque": "Queue",
}



##########################################################################################
# stub
##########################################################################################

from loguru import logger as log
from pathlib import Path
from typing import Union

import click
import stubber.utils as utils
from stubber.utils.post import do_post_processing

from .cli import stubber_cli

##########################################################################################
# log = logging.getLogger("stubber")
#########################################################################################


@stubber_cli.command(name="stub")
@click.option("--source", "-s", type=click.Path(exists=True, file_okay=True, dir_okay=True))
def cli_stub(source: Union[str, Path]):
    "Create or update .pyi type hint files."

    log.info("Generate type hint files (pyi) in folder: {}".format(source))
    OK = utils.generate_pyi_files(Path(source))
    do_post_processing([Path(source)], pyi=True, black=True)
    return 0 if OK else 1



"""
enrich machinestubs with docstubs
"""
from typing import List, Union

import click
from loguru import logger as log
from stubber.publish.merge_docstubs import merge_all_docstubs
from stubber.utils.config import CONFIG

from .cli import stubber_cli


@stubber_cli.command(name="merge")
@click.option("--family", default="micropython", type=str, show_default=True)
@click.option(
    "--version",
    "--Version",
    "-V",
    "versions",
    multiple=True,
    default=["auto"],
    # type=click.Choice(ALL_VERSIONS),
    show_default=True,
    help="'latest', 'auto', or one or more versions",
)
def cli_merge_docstubs(
    versions: Union[str, List[str]],
    family,
):
    """
    Enrich the stubs in stub_folder with the docstubs in docstubs_folder.
    """
    # single version should be a string
    if len(versions) == 1:
        versions = versions[0]
    log.info(f"Merge docstubs for {family} {versions}")
    _ = merge_all_docstubs(versions=versions, family=family, mpy_path=CONFIG.mpy_path)



# pragma: no cover
"""
generate the list of modules that should be attempted to stub
for this : 
- combine the modules from the different texts files
- split the lines into individual module names
- combine them in one set
- remove the ones than cannot be stubbed
- remove test modules, ending in `_test`
- write updates to:
    - board/modulelist.txt
    - board/createsubs.py 

- TODO: remove the frozen modules from this list
- TODO: bump patch number if there are actual changes

"""

from pathlib import Path
from typing import List, Optional, Set, Union

from loguru import logger as log


def read_modules(path: Optional[Path] = None) -> Set:
    """
    read text files with modules per firmware.
    each contains the output of help("modules")
    - lines starting with # are comments.
    - split the other lines at whitespace seperator,
    - and add each module to a set
    """
    path = Path(path or "./data")
    assert path
    all_modules = set()
    for file in path.glob("*.txt"):
        log.debug(f"processing: {file.name}")
        with file.open("r") as f:
            line = f.readline()
            while line:
                if len(line) > 1 and line[0] != "#":

                    file_mods = line.split()
                    log.trace(line[0:-1])
                    log.trace(set(file_mods))
                    # remove modules ending in _test
                    file_mods = [m for m in file_mods if not m.endswith("_test")]
                    all_modules = set(all_modules | set(file_mods))
                # next
                line = f.readline()

    log.trace(">" * 40)

    return all_modules


# def wrapped(modules: Union[Set, List]) -> str:
#     "wrap code line at spaces"
#     long_line = str(modules)
#     _wrapped = "        self.modules = "
#     IDENT = len(_wrapped)
#     MAX_WIDTH = 135

#     # find seperator
#     while len(long_line) > MAX_WIDTH:
#         p1 = long_line.find("', ", MAX_WIDTH)
#         # drop space
#         p1 += 3
#         short = long_line[0 : p1 - 1]
#         _wrapped += short + "\n" + " " * IDENT
#         long_line = long_line[p1 - 1 :]
#     _wrapped += long_line
#     return _wrapped


def main():
    """
    helper script
    generate a few lines of code with all modules to be stubbed by createstubs
    """
    #######################################################################
    # the exceptions
    #######################################################################
    mods_problematic = set(
        [
            "upysh",
            "webrepl_setup",
            "http_client",
            "http_client_ssl",
            "http_server",
            "http_server_ssl",
        ]
    )
    mods_excluded = set(
        [
            "__main__",
            "_main",
            "_boot",
            "webrepl",
            "_webrepl",
            "port_diag",
            "example_sub_led",
            "example_pub_button",
            "upip",
            "upip_utarfile",
            "upysh",
            # DOCSTUB_SKIP = [
            "uasyncio",  # can create better stubs from frozen python modules.
            "builtins",  # conflicts with static type checking , has very little information anyway
            "re",  # regex is too complex
            "collections",
            "io",
            "uio",
        ]
    )

    all_modules = read_modules()
    modules_to_stub = sorted(all_modules - set(mods_excluded | mods_problematic))

    # remove pycom MQTT* from defaults
    modules_to_stub = sorted({m for m in modules_to_stub if not m.startswith("MQTT")})

    log.info(f"modules to stub : {len(modules_to_stub)}")
    # log.debug(wrapped(modules_to_stub))
    lines: List[str] = []
    # update modules.txt
    for modules_txt in [Path("board/modulelist.txt"), Path("minified/modulelist.txt")]:
        if modules_txt.exists():
            with open(modules_txt) as f:
                lines = f.readlines()
                # only keep comment lines
                lines = [l for l in lines if l[0] == "#"]
        else:
            lines = ["# list of modules to stub."]
        lines = lines + [m + "\n" for m in modules_to_stub]
        log.info(f"writing module list to {modules_txt}")
        with open(modules_txt, "w") as f:
            f.writelines(lines)

    # update createstubs.py
    createstubs = Path("board/createstubs.py")
    if createstubs.exists():
        with open(createstubs) as f:
            lines = f.readlines()

    l_start = lines.index("    stubber.modules = [\n")
    assert l_start
    l_end = lines.index("    ]  # spell-checker: enable\n", l_start)
    assert l_end

    # Plug in the new list of modules
    lines = lines[: l_start + 1] + [f'        "{m}",\n' for m in modules_to_stub] + lines[l_end:]

    log.info(f"writing updated module list to {createstubs}")
    with open(createstubs, "w") as f:
        f.writelines(lines)


if __name__ == "__main__":
    main()



"""
get-lobo (frozen)
"""

from pathlib import Path
from typing import List

import click
import stubber.get_lobo as get_lobo
import stubber.utils as utils
from loguru import logger as log
from stubber.commands.cli import stubber_cli
from stubber.utils.config import CONFIG


@stubber_cli.command(name="get-lobo")
@click.option(
    "--stub-folder",
    "-stubs",
    default=CONFIG.stub_path.as_posix(),
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    show_default=True,
)
@click.option("--pyi/--no-pyi", default=True, help="Create .pyi files for the (new) frozen modules", show_default=True)
@click.option("--black/--no-black", default=True, help="Run black on the (new) frozen modules", show_default=True)
def cli_get_lobo(
    stub_folder: str = CONFIG.stub_path.as_posix(),
    pyi: bool = True,
    black: bool = True,
):
    """
    Get the frozen stubs for Lobo-esp32.

    """
    log.info("Get the frozen modules Loboris v3.2.24")

    stub_paths: List[Path] = []

    family = "loboris"
    version = "v3.2.24"
    stub_path = Path(stub_folder) / f"{family}-{utils.clean_version(version, flat=True)}-frozen"
    stub_paths.append(stub_path)
    get_lobo.get_frozen(stub_path)
    stub_paths = [stub_path]

    log.info(f"::group:: start post processing of retrieved stubs")
    utils.do_post_processing(stub_paths, pyi, black)
    log.info(f"::group:: Done")



from packaging.version import Version, parse


def bump_postrelease(
    current: Version,
    rc: int = 0,  # o, to bump post release, or release candidate number
) -> Version:
    """
    Increases the post release version number

    This allows for a new stub-release to be published while still using the Major.Minor.Patch version numbers of Micropython
    if rc = 0(default) : bump post release
        format: x.y.z.post1, x.y.z.post2 ...
    if rec specified:
        drop the post release and set the release candidate number

    ref: https://peps.python.org/pep-0440/
    """
    parts = []
    # Epoch
    if current.epoch != 0:
        parts.append(f"{current.epoch}!")
    # Release segment
    parts.append(".".join(str(x) for x in current.release))
    if rc != 0:
        # Pre-release / alpha / beta / rc segment
        parts.append(f".a{rc}")
    else:
        # BUMP Post-release
        if current.post is not None:
            parts.append(f".post{current.post + 1}")
        else:
            parts.append(f".post{1}")
    # Development release
    if current.dev is not None:
        parts.append(f".dev{current.dev}")

    # Local version segment
    if current.local is not None:
        parts.append(f"+{current.local}")

    new = parse("".join(parts))
    if not isinstance(new, Version):
        raise ValueError(f"{new} is not a valid version")

    return new



""" Merge firmware stubs and docstubs into a single folder
"""

import shutil
from pathlib import Path

from loguru import logger as log
from stubber.codemod.enrich import enrich_folder
from stubber.publish.candidates import firmware_candidates
from stubber.utils.config import CONFIG
from stubber.utils.versions import clean_version


def merge_all_docstubs(versions, family: str = "micropython", *, mpy_path=CONFIG.mpy_path):
    """merge docstubs into firmware stubs"""
    for fw in firmware_candidates(versions=versions, family=family):
        # check if we have firmware stubs of this version and port
        base = f"{fw['family']}-{clean_version(fw['version'],flat=True)}"
        fw_folder = base + f"-{fw['port']}"
        mrg_folder = fw_folder + "-merged"
        doc_folder = base + f"-docstubs"

        fw_path = CONFIG.stub_path / fw_folder
        mrg_path = CONFIG.stub_path / mrg_folder
        doc_path = CONFIG.stub_path / doc_folder

        if not fw_path.exists():
            # only continue if both folders exist
            continue
        if not doc_path.exists():
            print(f"Warning: no docstubs for {fw['version']}")
        log.info(f"Merge docstubs for {fw['family']} {fw['version']} {fw['port']} {fw['board']}")
        copy_docstubs(fw_path, mrg_path, doc_path)


def copy_docstubs(fw_path: Path, dest_path: Path, docstub_path: Path):
    """
    Parameters:
        fw_path: Path to firmware stubs (absolute path)
        dest_path: Path to destination (absolute path)
        mpy_version: micropython version ('1.18')

    Copy files from the firmware stub folders to the merged
    - 1 - Copy all firmware stubs to the package folder
    - 1.B - clean up a little bit
    - 2 - Enrich the firmware stubs with the document stubs

    """
    if dest_path.exists():
        # delete all files and folders from the destination
        shutil.rmtree(dest_path, ignore_errors=True)
    dest_path.mkdir(parents=True, exist_ok=True)

    # 1 - Copy  the stubs to the package, directly in the package folder (no folders)
    try:
        log.trace(f"Copying firmware stubs from {fw_path}")
        shutil.copytree(fw_path, dest_path, symlinks=True, dirs_exist_ok=True)
    except OSError as e:
        log.error(f"Error copying stubs from : { fw_path}, {e}")
        raise (e)
    # rename the module.json file to firmware.json
    if (dest_path / "modules.json").exists():
        (dest_path / "modules.json").rename(dest_path / "firmware_stubs.json")

    # 1.B - clean up a little bit
    do_cleanup = False
    if do_cleanup:
        # delete all the .py files in the package folder if there is a corresponding .pyi file
        # FIXME: Leave *.py on the module folders to avoid poetry packaging issues
        # >      ValueError  umqtt is not a package.

        for f in dest_path.glob("*.py"):
            if f.with_suffix(".pyi").exists():
                f.unlink()

    # delete buitins.pyi in the package folder
    for name in [
        "builtins.py",  # creates conflicts, better removed
        "builtins.pyi",  # creates conflicts, better removed
    ]:
        if (dest_path / name).exists():
            (dest_path / name).unlink()

    # 2 - Enrich the firmware stubs with the document stubs
    result = enrich_folder(dest_path, docstub_path=docstub_path, write_back=True)

    # copy the docstubs manifest.json file to the package folder
    # if (docstub_path / "modules.json").exists():
    shutil.copy(docstub_path / "modules.json", dest_path / "doc_stubs.json")
    return result



"""
Code from micropyton project and adapted to use the same versioning scheme
"""

from __future__ import print_function

import subprocess
from pathlib import Path

from packaging.version import parse


def get_version_info_from_git(path: Path = Path.cwd()):
    """return the version info from the git repository specified.
    returns: a 2-tuple containing git_tag, short_hash

    """
    # Python 2.6 doesn't have check_output, so check for that
    try:
        subprocess.check_output
        subprocess.check_call
    except AttributeError:  # pragma: no cover
        return None

    # Note: git describe doesn't work if no tag is available
    try:
        git_tag = subprocess.check_output(
            ["git", "describe", "--tags", "--dirty", "--always", "--match", "v[1-9].*"],
            stderr=subprocess.STDOUT,
            universal_newlines=True,
            cwd=path,
        ).strip()
    except subprocess.CalledProcessError as er:  # pragma: no cover
        if er.returncode == 128:
            # git exit code of 128 means no repository found
            return None
        git_tag = ""
    except OSError:
        return None
    try:
        git_hash = subprocess.check_output(
            ["git", "rev-parse", "--short", "HEAD"],
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        ).strip()
    except subprocess.CalledProcessError:  # pragma: no cover
        git_hash = "unknown"
    except OSError:
        return None

    try:
        # Check if there are any modified files.
        subprocess.check_call(["git", "diff", "--no-ext-diff", "--quiet", "--exit-code"], stderr=subprocess.STDOUT)
        # Check if there are any staged files.
        subprocess.check_call(["git", "diff-index", "--cached", "--quiet", "HEAD", "--"], stderr=subprocess.STDOUT)  # pragma: no cover
    except subprocess.CalledProcessError:
        git_hash += "-dirty"
    except OSError:
        return None

    return git_tag, git_hash


def get_version_build_from_git(path: Path = Path.cwd()):
    git_tag, short_hash = get_version_info_from_git(path)  # type: ignore
    assert git_tag is not None
    parts = git_tag.split("-")
    assert len(parts) >= 2
    ver = parse(parts[0])
    return ver, parts[1]


# def get_version_info_from_docs_conf():
#     with open(os.path.join(os.path.dirname(sys.argv[0]), "..", "docs", "conf.py")) as f:
#         for line in f:
#             if line.startswith("version = release = '"):
#                 ver = line.strip().split(" = ")[2].strip("'")
#                 git_tag = "v" + ver
#                 return git_tag, "<no hash>"
#     return None



"""
update-fallback folder with common set of stubs that cater for most of the devices
"""

from pathlib import Path

import click
from stubber.update_fallback import RELEASED, update_fallback
from stubber.update_module_list import main as update_module_list
from stubber.utils.config import CONFIG

from .cli import stubber_cli


@stubber_cli.command(name="update-module-list", help="Update the module list based on the information in the data folder")
def cli_update_module_list():
    """
    Update the module list based on the information in the data folder.
    """
    update_module_list()



import os
import shutil
from distutils.dir_util import copy_tree
from distutils.errors import DistutilsFileError
from pathlib import Path
from typing import List, Optional, Tuple

from loguru import logger as log

# log = logging.getLogger()

RELEASED = "v1_18"


def fallback_sources(version: str, fw_version: Optional[str] = None) -> List[Tuple[str, str]]:
    """
    list of sources to build/update the fallback 'catch-all' stubfolder
    version : the version to use
    fw_version : version to source the Firmware stubs from.  defaults to the version used , but can be lower

    """
    if not fw_version:
        fw_version = version
    if fw_version == "latest":
        fw_version = RELEASED
    SOURCES = [
        ("uasyncio", f"micropython-{fw_version}-esp32"),
        ("umqtt", f"micropython-{fw_version}-esp32"),
        ("_onewire.py*", f"micropython-{fw_version}-esp32"),
        ("_uasyncio.py*", f"micropython-{fw_version}-esp32"),
        ("array.py*", f"micropython-{fw_version}-esp32"),
        ("binascii.py*", f"micropython-{fw_version}-esp32"),
        ("hashlib.py*", f"micropython-{fw_version}-esp32"),
        ("machine.py*", f"micropython-{fw_version}-esp32"),
        ("micropython.py*", f"micropython-{version}-docstubs"),  # esp32"),
        ("network.pyi", f"micropython-{version}-docstubs"),  # esp32"),
        ("struct.py*", f"micropython-{fw_version}-esp32"),
        ("uarray.py*", f"micropython-{fw_version}-esp32"),
        ("ubinascii.py*", f"micropython-{fw_version}-esp32"),
        ("uctypes.py*", f"micropython-{fw_version}-esp32"),
        ("uerrno.py*", f"micropython-{fw_version}-esp32"),
        ("uhashlib.py*", f"micropython-{fw_version}-esp32"),
        ("uio.py*", f"micropython-{fw_version}-esp32"),
        ("ujson.py*", f"micropython-{fw_version}-esp32"),
        ("uselect.py*", f"micropython-{fw_version}-esp32"),
        ("usocket.py*", f"micropython-{fw_version}-esp32"),
        ("ussl.py*", f"micropython-{fw_version}-esp32"),
        ("ustruct.py*", f"micropython-{fw_version}-esp32"),
        ("sys.py*", f"micropython-{fw_version}-esp32"),
        ("usys.py*", f"micropython-{fw_version}-esp32"),
        ("uzlib.py*", f"micropython-{fw_version}-esp32"),
        ("bluetooth.py*", f"micropython-{version}-docstubs"),
        # esp
        ("esp.py*", f"micropython-{version}-docstubs"),  # 8266 fw stub has most functions
        ("esp32.py*", f"micropython-{version}-docstubs"),  # esp32"),
        # pyboard == stm32
        ("pyb.py*", f"micropython-{fw_version}-stm32"),
        ("framebuf.py*", f"micropython-{fw_version}-stm32"),
        # rp2
        ("_rp2.py*", f"micropython-{fw_version}-rp2"),
        # stdlib
        ("os.py*", f"micropython-{fw_version}-esp32"),  # -> stdlib
        ("uos.py*", f"micropython-{fw_version}-esp32"),
        ("time.py*", f"micropython-{fw_version}-esp32"),  # -> stdlib
        ("utime.py*", f"micropython-{fw_version}-esp32"),
    ]
    return SOURCES


def update_fallback(stubpath: Path, fallback_path: Path, version: str = RELEASED):
    "update the fallback stubs from the defined sources"
    # remove all *.py/.pyi files
    if not fallback_path.exists():
        os.makedirs(fallback_path)
    else:
        oldstubs = list(fallback_path.rglob("*.py")) + list(fallback_path.rglob("*.pyi"))
        log.debug(f"deleting {len(oldstubs)} stubs from {fallback_path.as_posix()}")
        for f in oldstubs:
            try:
                os.remove(f)
            except OSError as e:
                log.warning(e)
    added = 0
    for (name, source) in fallback_sources(version):
        if not "." in name:
            # copy folder
            log.debug(f"add {source} folder")
            try:
                copy_tree(
                    (stubpath / source / name).as_posix(),
                    (fallback_path / name).as_posix(),
                )
                added += 1
            except DistutilsFileError:
                log.warning(f"{(stubpath / source / name).as_posix()} not found")
        else:
            # copy file(s)
            log.debug(f"add {source}")
            for f in (stubpath / source).glob(name):
                shutil.copyfile(f, fallback_path / f.name)
                added += 1
    return added


if __name__ == "__main__":
    from stubber.utils.config import CONFIG

    update_fallback(
        CONFIG.stub_path,
        CONFIG.fallback_path,
        version=RELEASED,
    )



"""
switch 
"""

import os
from pathlib import Path
from typing import Optional, Union

import click
import stubber.basicgit as git
from loguru import logger as log
from stubber.utils.config import CONFIG
from stubber.utils.repos import match_lib_with_mpy

from .cli import stubber_cli

##########################################################################################
# log = logging.getLogger("stubber")
#########################################################################################


# get version list from Git tags in the repo that is provided on the command line
try:
    VERSION_LIST = git.get_tags(CONFIG.mpy_path, minver="v1.10") + ["v1.9.3", "v1.9.4", "latest"]
except Exception:
    VERSION_LIST = ["latest"]  # type: ignore


@stubber_cli.command(name="switch")
# WIP: Transition from Option to  Argument, currently supports both
@click.argument("tag", required=False, type=click.Choice(VERSION_LIST, case_sensitive=False))
@click.option("--path", "-p", default=CONFIG.repo_path.as_posix(), type=click.Path(file_okay=False, dir_okay=True))
def cli_switch(path: Union[str, Path], tag: Optional[str] = None):
    """
    Switch to a specific version of the micropython repos.

    Specify the version of the MicroPython repo.
    use of the --tag or --version is deprocated

    The Micropython-lib repo will be checked out to a commit that corresponds
    in time to that version tag, in order to allow non-current versions to be
    stubbed correctly.

    The repros must be cloned already
    """
    dest_path = Path(path)
    if not dest_path.exists():
        os.mkdir(dest_path)
    # repos are relative to provided path
    if dest_path != CONFIG.repo_path:
        mpy_path = dest_path / "micropython"
        mpy_lib_path = dest_path / "micropython-lib"
    else:
        mpy_path = CONFIG.mpy_path
        mpy_lib_path = CONFIG.mpy_lib_path

    # if no repos then error
    if not (mpy_path / ".git").exists():
        log.error("micropython repo not found")
        return -1
    if not (mpy_lib_path / ".git").exists():
        log.error("micropython-lib repo not found")
        return -1

    # fetch then switch
    log.info(f"fetch updates")
    git.fetch(mpy_path)
    git.fetch(mpy_lib_path)
    try:
        git.fetch(CONFIG.stub_path.parent)
    except Exception:
        log.trace("no stubs repo found : {CONFIG.stub_path.parent}")

    if not tag or tag == "":
        tag = "latest"

    log.info(f"Switching to {tag}")
    if tag == "latest":
        git.switch_branch(repo=mpy_path, branch="master")
    else:
        git.checkout_tag(repo=mpy_path, tag=tag)
    match_lib_with_mpy(version_tag=tag, lib_path=mpy_lib_path)

    log.info(f"{mpy_lib_path} {git.get_tag(mpy_path)}")
    log.info(f"{mpy_lib_path} {git.get_tag(mpy_lib_path)}")
    return 0



import re
from pathlib import Path
from typing import Any, Dict, Generator, List, Union

import stubber.basicgit as git
from packaging.version import parse
from stubber.publish.enums import COMBO_STUBS, DOC_STUBS, FIRMWARE_STUBS
from stubber.utils.config import CONFIG
from stubber.utils.versions import clean_version, micropython_versions

OLDEST_VERSION = "1.16"
"This is the oldest MicroPython version to build the stubs on"

V_LATEST = "latest"


def subfolder_names(path: Path):
    "returns a list of names of the subfolders of the given path"
    if path.exists():
        for child in path.iterdir():
            if child.is_dir():
                yield child.name


def version_cadidates(suffix: str, prefix=r".*", *, path=CONFIG.stub_path, oldest=OLDEST_VERSION) -> Generator[str, None, None]:
    "get a list of versions for the given family and suffix"
    if path.exists():
        folder_re = prefix + "-(.*)-" + suffix
        for name in subfolder_names(path):
            match = re.match(folder_re, name)
            if match:
                folder_ver = clean_version(match.group(1))
                if folder_ver == V_LATEST or parse(folder_ver) >= parse(oldest):
                    yield folder_ver


def list_frozen_ports(
    family: str = "micropython",
    version: str = V_LATEST,
    path=CONFIG.stub_path,
):
    "get list of ports with frozen stubs for a given family and version"
    ports_path = path / f"{family}-{version}-frozen"
    ports = list(subfolder_names(ports_path))
    return ports


def list_micropython_ports(
    family: str = "micropython",
    version: str = V_LATEST[0],
    mpy_path=CONFIG.mpy_path,
):
    "get list of micropython ports for a given family and version"
    if family != "micropython":
        # todo: add support for other families
        return []
    mpy_path = Path("./repos/micropython")
    # check out the micropthon repo for this version
    if not git.checkout_tag(version, mpy_path):
        return []
    ports_path = mpy_path / "ports"
    ports = list(subfolder_names(ports_path))
    # remove blocked ports from list
    for port in ["minimal", "bare-arm"]:  # CONFIG.blocked_ports:
        if port in ports:
            ports.remove(port)
    return ports


def frozen_candidates(
    family: str = "micropython",
    versions: Union[str, List[str]] = V_LATEST,
    ports: Union[str, List[str]] = "auto",
    boards: Union[str, List[str]] = "auto",
    *,
    path=CONFIG.stub_path,
) -> Generator[Dict[str, Any], None, None]:
    """
    generate a list of possible firmware stubs for the given family (, version port and board) ?
    - family = micropython
        board and port are ignored, they are looked up from the available frozen stubs
    - versions = 'latest' , 'auto' or a list of versions
    - port = 'auto' or a specific port
    - board = 'auto' or a specific board, 'GENERIC' shoould be specifid in CAPS
    """
    auto_port = isinstance(ports, str) and "auto" == ports or isinstance(ports, list) and "auto" in ports
    auto_board = isinstance(boards, str) and "auto" == boards or isinstance(boards, list) and "auto" in boards
    auto_version = isinstance(versions, str) and "auto" == versions or isinstance(versions, list) and "auto" in versions

    if isinstance(versions, str):
        if auto_version:
            versions = list(version_cadidates(suffix="frozen", prefix=family, path=path)) + [V_LATEST]
        else:
            versions = [versions]
    versions = [clean_version(v, flat=True) for v in versions]

    if isinstance(ports, str):
        ports = [ports]
    if isinstance(boards, str):
        boards = [boards]
    # ---------------------------------------------------------------------------
    # ---------------------------------------------------------------------------
    for version in versions:
        if auto_port:
            if family == "micropython":
                # lookup the (frozen) micropython ports
                ports = list_frozen_ports(family, version, path=path)
            else:
                raise NotImplementedError(f"auto ports not implemented for family {family}")  # pragma: no cover
            # elif family == "pycom":
            #     ports = ["esp32"]
            # elif family == "lobo":
            #     ports = ["esp32"]
        # ---------------------------------------------------------------------------
        for port in ports:
            boards_path = path / f"{family}-{version}-frozen" / port
            if boards_path.exists():
                yield {"family": family, "version": version, "port": port, "board": "GENERIC", "pkg_type": COMBO_STUBS}
            if auto_board:
                if family == "micropython":
                    # lookup the (frozen) micropython ports
                    boards = list(subfolder_names(boards_path))
                    # TODO: remove non-relevant boards
                    # - release - used in release testing
                    # generic_512 - small memory footprint
                    #

                else:
                    # raise NotImplementedError(f"auto boards not implemented for family {family}")  # pragma: no cover
                    raise Exception(f"auto boards not implemented for family {family}")  # pragma: no cover
                # elif family == "pycom":
                #     boards = ["wipy", "lopy", "gpy", "fipy"]
            # ---------------------------------------------------------------------------
            for board in boards:
                assert isinstance(board, str)
                # prozen stubs found , and not excluded, generic is already explicitly included
                if (path / f"{family}-{version}-frozen" / port / board).exists() and board.upper() not in [
                    "GENERIC",
                    "RELEASE",
                    "GENERIC_512K",
                ]:
                    yield {"family": family, "version": version, "port": port, "board": board, "pkg_type": COMBO_STUBS}


def docstub_candidates(
    family: str = "micropython",
    versions: Union[str, List[str]] = V_LATEST,
    path=CONFIG.stub_path,
):
    """generate a list of possible documentation stub candidates for the given family and version.

    Note that the folders do not need to exist, with the exeption of auto which will scan the stubs folder for versions of docstubs
    """
    if isinstance(versions, str):
        if "auto" == versions:  # auto with vprefix ...
            versions = list(version_cadidates(suffix="docstubs", prefix=family, path=path))
        else:
            versions = [versions]
    versions = [clean_version(v, flat=True) for v in versions]

    for version in versions:
        yield {"family": family, "version": version, "pkg_type": DOC_STUBS}


def firmware_candidates(
    family: str = "micropython",
    versions: Union[str, List[str]] = V_LATEST,
    *,
    mpy_path=CONFIG.mpy_path,
):
    """generate a list of possible firmware stub candidates for the given family and version."""
    if isinstance(versions, str):
        if "auto" == versions:  # auto with vprefix ...
            versions = list(micropython_versions(start=OLDEST_VERSION))
        else:
            versions = [versions]
    versions = [clean_version(v, flat=False) for v in versions]

    for version in versions:
        ports = list_micropython_ports(family, version, mpy_path=mpy_path)
        for port in ports:
            yield {"family": family, "version": version, "port": port, "board": "", "pkg_type": FIRMWARE_STUBS}



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
import argparse
from loguru import logger as log
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import libcst as cst
from libcst.codemod import CodemodContext, VisitorBasedCodemodCommand
from libcst.codemod.visitors import AddImportsVisitor, GatherImportsVisitor, ImportItem
from stubber.cst_transformer import MODULE_KEY, StubTypingCollector, TypeInfo, update_def_docstr, update_module_docstr

##########################################################################################
# # log = logging.getLogger(__name__)
#########################################################################################


class MergeCommand(VisitorBasedCodemodCommand):
    """
    A libcst transformer that merges the type-rich information from a doc-stub into
    a firmware stub.
    The resulting file will contain information from both sources.

    - module docstring - from source

    - function parameters and types - from docstubs
    - function return types - from docstubs
    - function docstrings - from source

    """

    DESCRIPTION: str = "Merge the type-rich information from a doc-stub into a firmware stub"

    @staticmethod
    def add_args(arg_parser: argparse.ArgumentParser) -> None:
        """Add command-line args that a user can specify for running this codemod."""

        arg_parser.add_argument(
            # "-sf",
            "--stubfile",
            dest="stub_file",
            metavar="PATH",
            help="The path to the doc-stub file",
            type=str,
            required=True,
        )

    def __init__(self, context: CodemodContext, stub_file: Union[Path, str]) -> None:
        super().__init__(context)
        self.replace_functiondef_with_classdef = True
        # stack for storing the canonical name of the current function/method
        self.stack: List[str] = []
        # stubfile is the path to the doc-stub file
        self.stub_path = Path(stub_file)
        # read the stub file from the path
        self.stub_source = self.stub_path.read_text(encoding="utf-8")
        # store the annotations
        self.annotations: Dict[
            Tuple[str, ...],  # key: tuple of canonical class/function name
            TypeInfo,  # value: TypeInfo
        ] = {}

        self.stub_imports: Dict[str, ImportItem] = {}
        # parse the doc-stub file
        if self.stub_source:
            try:
                # parse the doc-stub file
                stub_tree = cst.parse_module(self.stub_source)
            except cst.ParserSyntaxError as e:
                log.error(f"Error parsing {self.stub_path}: {e}")
                return
            # create the collectors
            typing_collector = StubTypingCollector()
            import_collector = GatherImportsVisitor(context)
            # visit the doc-stub file with all collectors
            stub_tree.visit(typing_collector)
            self.annotations = typing_collector.annotations

            # Store the imports that were added to the stub file
            stub_tree.visit(import_collector)
            self.stub_imports = import_collector.symbol_mapping

    # ------------------------------------------------------------------------

    def leave_Module(self, original_node: cst.Module, updated_node: cst.Module) -> cst.Module:
        "Update the Module docstring"
        # add any needed imports from the doc-stub
        for k in self.stub_imports.keys():
            _imp = self.stub_imports[k]
            log.trace(f"import {k} = {_imp}")
            AddImportsVisitor.add_needed_import(
                self.context,
                module=_imp.module_name,
                obj=_imp.obj_name,
                asname=_imp.alias,
                relative=_imp.relative,
            )

        # update the docstring.
        if not MODULE_KEY in self.annotations:
            # no changes
            return updated_node

        # update/replace  module docstrings
        # todo: or should we add / merge the docstrings?
        new = self.annotations[MODULE_KEY]

        return update_module_docstr(updated_node, new.docstr_node)

    # ------------------------------------------------------------
    #  keep track of the the (class, method) names to the stack
    def visit_ClassDef(self, node: cst.ClassDef) -> Optional[bool]:
        self.stack.append(node.name.value)

    def leave_ClassDef(self, original_node: cst.ClassDef, updated_node: cst.ClassDef) -> cst.ClassDef:
        stack_id = tuple(self.stack)
        self.stack.pop()
        if not stack_id in self.annotations:
            # no changes to the function
            return updated_node
        # update the firmware_stub from the doc_stub information
        new = self.annotations[stack_id]
        # first update the docstring
        updated_node = update_def_docstr(updated_node, new.docstr_node)
        # Sometimes the firmware stubs and the doc stubs have different types : FunctionDef / ClassDef
        # we need to be carefull not to copy over all the annotations if the types are different
        if new.def_type == "classdef":
            # Same type, we can copy over all the annotations
            return updated_node.with_changes(decorators=new.decorators)
        elif new.def_type == "funcdef":
            # Different type: ClassDef != FuncDef ,
            # for now just return the updated node
            return updated_node
        else:
            #  just return the updated node
            return updated_node

    # ------------------------------------------------------------------------
    def visit_FunctionDef(self, node: cst.FunctionDef) -> Optional[bool]:
        self.stack.append(node.name.value)
        return True

    def leave_FunctionDef(self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef) -> Union[cst.FunctionDef, cst.ClassDef]:
        "Update the function Parameters and return type, decorators and docstring"

        stack_id = tuple(self.stack)
        self.stack.pop()
        if not stack_id in self.annotations:
            # no changes to the function
            return updated_node
        # update the firmware_stub from the doc_stub information
        new = self.annotations[stack_id]

        # first update the docstring
        updated_node = update_def_docstr(updated_node, new.docstr_node, new.def_node)
        # Sometimes the firmware stubs and the doc stubs have different types : FunctionDef / ClassDef
        # we need to be carefull not to copy over all the annotations if the types are different
        if new.def_type == "funcdef":
            # Same type, we can copy over all the annotations
            return updated_node.with_changes(
                params=new.params,
                returns=new.returns,
                decorators=new.decorators,
            )
        elif new.def_type == "classdef":
            # Different type: ClassDef != FuncDef ,
            if new.def_node and self.replace_functiondef_with_classdef:
                # replace the functiondef with the classdef from the stub file
                return new.def_node
            # for now just return the updated node
            return updated_node
        else:
            #  just return the updated node
            return updated_node



"""
simple Git module, where needed via powershell
"""
import subprocess
from pathlib import Path
from typing import List, Optional, Union

from loguru import logger as log
from packaging.version import parse


def _run_git(
    cmd: List[str],
    repo: Optional[Union[Path, str]] = None,
    expect_stderr=False,
    capture_output=True,
    echo_output=True,
):
    "run a external (git) command in the repo's folder and deal with some of the errors"
    try:
        if repo:
            if isinstance(repo, str):
                repo = Path(repo)
            result = subprocess.run(cmd, capture_output=capture_output, check=True, cwd=repo.absolute().as_posix())
        else:
            result = subprocess.run(cmd, capture_output=capture_output, check=True)
    except (NotADirectoryError, FileNotFoundError) as e:  # pragma: no cover
        return None
    except subprocess.CalledProcessError as e:  # pragma: no cover
        # add some logging for github actions
        log.error("Exception on process, rc=", e.returncode, "output=", e.output.decode("utf-8"), "\nstderr=", e.stderr.decode("utf-8"))
        return None
    if result.stderr and result.stderr != b"":
        stderr = result.stderr.decode("utf-8")
        if "warning" in stderr.lower():
            log.warning(stderr)
            expect_stderr = True
        elif capture_output and echo_output:  # pragma: no cover
            log.error(stderr)
        if not expect_stderr:
            raise Exception(stderr)

    if result.returncode < 0:
        raise Exception(result.stderr.decode("utf-8"))
    return result


def clone(remote_repo: str, path: Path, shallow=False, tag: Optional[str] = None) -> bool:
    "git clone [--depth 1] [--branch <tag_name>] <remote> <directory>"
    cmd = ["git", "clone"]
    if shallow:
        cmd += ["--depth", "1"]
    if tag in ("latest", "master"):
        tag = None
    if not tag:
        cmd += [remote_repo, str(path)]
    else:
        cmd += [remote_repo, "--branch", tag, str(path)]
    result = _run_git(cmd, expect_stderr=True, capture_output=False)
    if result:
        return result.returncode == 0
    else:
        return False


def get_tag(repo: Optional[Union[str, Path]] = None, abbreviate: bool = True) -> Union[str, None]:
    """
    get the most recent git version tag of a local repo
    repo Path should be in the form of : repo = "./repo/micropython"

    returns the tag or None
    """
    if not repo:
        repo = Path(".")
    elif isinstance(repo, str):
        repo = Path(repo)

    result = _run_git(["git", "describe"], repo=repo.as_posix(), expect_stderr=True)
    if not result:
        return None
    tag: str = result.stdout.decode("utf-8")
    tag = tag.replace("\r", "").replace("\n", "")
    if abbreviate and "-" in tag:
        # this may or not be the latest on the main branch
        # result = _run_git(["git", "rev-parse", "--abbrev-ref", "HEAD"], repo=repo, expect_stderr=True)
        result = _run_git(["git", "status", "--branch"], repo=repo.as_posix(), expect_stderr=True)
        if result:
            lines = result.stdout.decode("utf-8").replace("\r", "").split("\n")
            if lines[0].startswith("On branch"):
                if lines[0].endswith("master"):
                    tag = "latest"
                elif lines[0].endswith("fix_lib_documentation"):
                    tag = "fix_lib_documentation"
    return tag


def get_tags(repo: Optional[Path] = None, minver: Optional[str] = None) -> List[str]:
    """
    get list of tag of a local repo
    """
    if not repo:
        repo = Path(".")

    result = _run_git(["git", "tag", "-l"], repo=repo.as_posix(), expect_stderr=True)
    if not result or result.returncode != 0:
        return []
    tags = result.stdout.decode("utf-8").replace("\r", "").split("\n")
    tags = [tag for tag in tags if tag.startswith("v")]
    if minver:
        tags = [tag for tag in tags if parse(tag) >= parse(minver)]
    return sorted(tags)


def checkout_tag(tag: str, repo: Optional[Union[str, Path]] = None) -> bool:
    """
    checkout a specific git tag
    """
    cmd = ["git", "checkout", "tags/" + tag, "--detach", "--quiet", "--force"]
    result = _run_git(cmd, repo=repo, expect_stderr=True, capture_output=True)
    if not result:
        return False
    # actually a good result
    log.debug(result.stderr.decode("utf-8"))
    log.debug(result.stdout.decode("utf-8"))
    return True


def synch_submodules(repo: Optional[Union[Path, str]] = None) -> bool:
    """
    make sure any submodules are in syncj
    """
    cmds = [
        ["git", "submodule", "sync", "--quiet"],
        ["git", "submodule", "update", "--quiet"],
    ]
    for cmd in cmds:
        result = _run_git(cmd, repo=repo, expect_stderr=True)
        if not result:
            return False
        # actually a good result
        log.debug(result.stderr.decode("utf-8"))
    return True


def checkout_commit(commit_hash: str, repo: Optional[Union[Path, str]] = None) -> bool:
    """
    Checkout a specific commit
    """
    cmd = ["git", "checkout", commit_hash, "--quiet", "--force"]
    result = _run_git(cmd, repo=repo, expect_stderr=True)
    if not result:
        return False
    # actually a good result
    log.debug(result.stderr.decode("utf-8"))
    synch_submodules(repo)
    return True


def switch_tag(tag: str, repo: Optional[Union[Path, str]] = None) -> bool:
    """
    get the most recent git version tag of a local repo"
    repo should be in the form of : path/.git
    repo = '../micropython/.git'
    returns the tag or None
    """

    cmd = ["git", "switch", "--detach", tag, "--quiet", "--force"]
    result = _run_git(cmd, repo=repo, expect_stderr=True)
    if not result:
        return False
    # actually a good result
    log.debug(result.stderr.decode("utf-8"))
    synch_submodules(repo)
    return True


def switch_branch(branch: str, repo: Optional[Union[Path, str]] = None) -> bool:
    """
    get the most recent git version tag of a local repo"
    repo should be in the form of : path/.git
    repo = '../micropython/.git'
    returns the tag or None
    """
    cmd = ["git", "switch", branch, "--quiet", "--force"]
    result = _run_git(cmd, repo=repo, expect_stderr=True)
    if not result:
        return False
    # actually a good result
    log.debug(result.stderr.decode("utf-8"))
    synch_submodules(repo)
    return True


def fetch(repo: Union[Path, str]) -> bool:
    """
    fetches a repo
    repo should be in the form of : path/.git
    repo = '../micropython/.git'
    returns True on success
    """
    if not repo:
        raise NotADirectoryError

    cmd = ["git", "fetch", "--all", "--tags", "--quiet"]
    result = _run_git(cmd, repo=repo, echo_output=False)
    if not result:
        return False
    return result.returncode == 0


def pull(repo: Union[Path, str], branch="main") -> bool:
    """
    pull a repo origin into main
    repo should be in the form of : path/.git
    repo = '../micropython/.git'
    returns True on success
    """
    if not repo:
        raise NotADirectoryError
    repo = Path(repo)
    # first checkout HEAD
    cmd = ["git", "checkout", branch, "--quiet", "--force"]
    result = _run_git(cmd, repo=repo, expect_stderr=True)
    if not result:
        log.error("error during git checkout main", result)
        return False

    cmd = ["git", "pull", "origin", branch, "--quiet", "--autostash"]
    result = _run_git(cmd, repo=repo, expect_stderr=True)
    if not result:
        log.error("error durign pull", result)
        return False
    return result.returncode == 0
